from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from a99 import WDBRegistry
import a99
from .a_WDBFCF import *
from .a_WDBPFANTMol import *
from .a_WDBSystem import *

__all__ = ["WDBSystemPFANTMolFCF"]


class WDBSystemPFANTMolFCF(a99.WBase):
    """Master-slave widget to lodge System and FCF widgets"""

    # Emitted whenever the value of the 'id' property changes
    id_changed = pyqtSignal()

    @property
    def id(self):
        """System id or None"""
        return self.w_system._get_id()

    @property
    def f(self):
        """Object representing the file being edited (possibly a DataFile object)"""
        return self._f

    @f.setter
    def f(self, f):
        self.w_system.f = f
        self.w_fcf.f = f
        self.w_pfantmol.f = f

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        # # Main layout & splitter
        lmain = self.keep_ref(QVBoxLayout(self))
        sp = self.keep_ref(QSplitter(Qt.Horizontal))


        # ## First widget of splitter
        w0 = self.keep_ref(QWidget())
        l0 = QVBoxLayout(w0)
        a99.set_margin(l0, 0)
        l0.setSpacing(2)

        a = self.title_mol = QLabel(a99.format_title1("Systems"))
        l0.addWidget(a)

        # ### label and edit with copiable system description
        a = self.label_sysdescr = QLabel("System descriptio&n:")
        b = self.lineedit_sysdescr = QLineEdit()
        b.setReadOnly(True)
        b.setFixedWidth(200)
        a.setBuddy(b)

        l = self.layout_sysdescr = QHBoxLayout()
        l.addWidget(a)
        l.addWidget(b)
        l.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        l0.addLayout(l)

        # ### Systems registry table
        w = self.w_system = WDBSystem(self.parent_form)
        w.id_changed.connect(self.id_system_changed)
        w.id_changed.connect(self.id_changed)
        w.changed.connect(self.changed)
        l0.addWidget(w)

        # ## Second widget of splitter
        sp1 = self.keep_ref(QSplitter(Qt.Vertical))

        # ### widget 1.1

        w10 = self.keep_ref(QWidget())
        l10 = QVBoxLayout(w10)
        a99.set_margin(l10, 0)
        l10.setSpacing(2)

        a = self.title_state = self.keep_ref(QLabel(a99.format_title1("PFANT molecules")))
        l10.addWidget(a)

        w = self.w_pfantmol = WDBPFANTMol(self.parent_form)
        w.changed.connect(self.changed)
        l10.addWidget(w)

        # ### widget 1.1

        w11 = self.keep_ref(QWidget())
        l11 = QVBoxLayout(w11)
        a99.set_margin(l11, 0)
        l11.setSpacing(2)

        a = self.title_state = self.keep_ref(QLabel(a99.format_title1("Franck-Condon Factors")))
        l11.addWidget(a)

        w = self.w_fcf = WDBFCF(self.parent_form)
        w.changed.connect(self.changed)
        l11.addWidget(w)


        # ## Adds widgets to splitters

        sp1.addWidget(w10)
        sp1.addWidget(w11)

        sp.addWidget(w0)
        sp.addWidget(sp1)

        lmain.addWidget(sp)

        # # Final adjustments
        a99.nerdify(self)

        self.w_system._move_to_first()


    def set_id_molecule(self, id_):
        """Propagates to self.w_system"""
        self.w_system.set_id_molecule(id_)
        # self._populate()

    def id_system_changed(self):
        import pyfant
        self.w_pfantmol.set_id_system(self.w_system.id)
        self.w_fcf.set_id_system(self.w_system.id)

        row = self.w_system.row
        descr = pyfant.molconsts_to_system_str(row, pyfant.SS_PLAIN) if row else ""

        wfilemoldb = self.parent().parent().parent().parent()
        molrow = wfilemoldb.w_mol.row

        descr_ = "" if not molrow else f"{molrow['formula']} [{descr}]"

        self.lineedit_sysdescr.setText(descr_)




from collections import Counter
import a99, tabulate

@a99.froze_it
class MolConversionLog(object):

    @property
    def num_lines_skipped(self):
        """Sums all counts in self.skip_reasons"""
        return sum(self.skip_reasons.values())

    @property
    def n(self):
        return self.num_lines

    @n.setter
    def n(self, x):
        self.num_lines = x

    def __init__(self, num_lines=0, flag_ok=True):
        # Number of lines in input file
        self.num_lines = num_lines
        # Whether the conversion was considered successful (despite errors)
        self.flag_ok = flag_ok
        # List of error strings
        self.errors = []
        # Counter: {"reason": number_of_times, ...}
        # The key should be a descriptive reason, however shortish
        # Access like this:
        #     >>> log.skip_reasons["FCF not found for (vl,v2l)"] += 1
        #
        self.skip_reasons = Counter()

        self.cnt_in = 0

    def __str__(self):
        INDENT = "    "
        n = self.num_lines
        _ret = [f"num_lines: {n}", f"flag_ok: {self.flag_ok}", f"included: {self.cnt_in}/{n}", f"excluded: {n-self.cnt_in}/{n}"]

        if self.skip_reasons:
            _ret.append("\n*** Skip reasons ***")
            _ret.append("\n".join([INDENT+line for line in tabulate.tabulate([(k, v) for k, v in self.skip_reasons.items()], ["Reason", "number of occurences"]).split("\n")]))
        if self.errors:
            _ret.append("\n*** Errors ***")
            _ret.append("\n".join([INDENT+line for line in self.errors]))
        return "\n".join(_ret)



__all__ = ["XRunnableManager"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from threading import Lock
import time
import matplotlib.pyplot as plt
import a99
import pyfant
import f311


COLORS = [QColor(255, 0, 0),
          QColor(192, 0, 0),
          QColor(96, 96, 96),
          QColor(128, 128, 128)]
STATUS_COLUMN_WIDTH = 730
WINDOW_WIDTH = 700
WINDOW_HEIGHT = 700
# # Defaults for checkable buttons (True: checked)
FLAG_SHOW_STATUS = True
FLAG_SHOW_RUNNABLES = True


class XRunnableManager(QMainWindow):
    """
    Thread manager window.

    Allows to monitor running runnables.
    """

    def __init__(self, parent=None, rm=None):
        QMainWindow.__init__(self, parent)
        assert isinstance(rm, pyfant.RunnableManager)
        self.rm = rm
        # Whether to close matplotlib plots when window is closed
        self.flag_close_mpl_plots_on_close = True
        # Whether to show "close window to continue" when finished
        self.flag_close_message = True
        self.__num_finished_shown = 0  # outdated version of self.rm.num_finished
        self.__lock_table = Lock()
        self.__lfcs = 0  # labelFinished color state
        self.__explorer_form = None


        # # Window design

        # ## Toolbar

        b0 = self.pushButtonStatus = QPushButton("Show &status")
        b0.clicked.connect(self.on_status)
        b0.setCheckable(True)
        b0.setChecked(FLAG_SHOW_STATUS)
        b0.setToolTip("Turn status panel on/off")
        b1 = self.pushButtonTable = QPushButton("Show &runnables")
        b1.clicked.connect(self.on_table)
        b1.setCheckable(True)
        b1.setChecked(FLAG_SHOW_RUNNABLES)
        b1.setToolTip("Turn status panel on/off")
        b2 = self.pushButtonPause = QPushButton("&Pause")
        b2.clicked.connect(self.on_pause)
        b2.setCheckable(True)
        b2.setToolTip("Pause/Resume")
        b3 = self.pushButtonCancel = QPushButton("&Kill running")
        b3.setToolTip("Attempts to kill all running jobs")
        b3.clicked.connect(self.on_kill)
        b4 = self.pushButtonCancel = QPushButton("&Exit thread")
        b4.setToolTip("Stops runnables manager thread (irreversible)")
        b4.clicked.connect(self.on_exit)
        b5 = self.pushButtonRetryFailed = QPushButton("Retry &failed")
        b5.clicked.connect(self.on_retry_failed)
        b6 = self.pushButtonCollectErrors = QPushButton("&Collect errors")
        b6.setToolTip("Searches for errors in log files and reports these errors in a new window.")
        b6.clicked.connect(self.on_collect_errors)
        # todo + note: assuming that there is no interest in adding new tasks later
        l = self.labelEnd = QLabel()
        l.setVisible(False)

        s = self.spacer0 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        bb = [b0, b1, b2, b3, b4, b5, b6, l]
        l0 = self.layoutToolbar = QHBoxLayout()
        for b in bb:
            l0.addWidget(b)
        l0.addItem(s)

        a = self.widgetToolbar = QWidget()
        a.setLayout(l0)
        a.setFixedHeight(40)

        # ## Status widget

        # ### Left column
        w0 = self.labelStatusFinished = QLabel()
        w1 = self.labelStatusFailed = QLabel()
        w2 = self.labelStatusElla = QLabel()
        w3 = self.labelStatusTPR = QLabel()
        w4 = self.labelStatusTotal = QLabel()
        w5 = self.labelStatusRema = QLabel()
        ww = [("Num finished", w0),
              ("Num failed", w1),
              ("Time ellapsed", w2),
              ("Time total estimate", w4),
              ("Time remaining estimate", w5),
              ("Average time per runnable", w3)]
        l = self.layoutStatus0 = QFormLayout()
        for caption, w in ww:
            w.setStyleSheet("QLabel {font-weight: bold}")
            l.addRow(caption, w)

        # ### Right column
        w0 = self.checkBox_paused = QCheckBox()
        w1 = self.checkBox_cancelled = QCheckBox()
        w2 = self.checkBox_finished = QCheckBox()
        w3 = self.checkBox_failed = QCheckBox()
        w4 = self.checkBox_exited = QCheckBox()
        l = self.layoutStatus1 = QFormLayout()
        ww = [("Paused", w0),
              ("Cancelled", w1),
              ("Finished", w2),
              ("Failed", w3),
              ("Exited", w4)]
        for caption, w in ww:
            w.setChecked(False)
            w.setEnabled(False)
            l.addRow(caption, w)

        l = QHBoxLayout()
        l.addLayout(self.layoutStatus0)
        l.addLayout(self.layoutStatus1)
        x = self.widgetStatus = QWidget()
        x.setVisible(b0.isChecked())
        x.setLayout(l)


        # ## Table Widget

        a = self.tableWidget = QTableWidget()
        a.setToolTip("Double-click to open session directory")
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setAlternatingRowColors(True)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setSelectionBehavior(QTableWidget.SelectRows)
        a.cellDoubleClicked.connect(self.on_tableWidget_cellDoubleClicked)
        # a.setFont(MONO_FONT)
        a.setVisible(self.pushButtonTable.isChecked())
        a.installEventFilter(self)

        # ## Mounts central widget

        l2 = self.layoutCentral = QVBoxLayout()
        ww = [self.widgetToolbar, self.widgetStatus, self.tableWidget]
        for w in ww:
            l2.addWidget(w)
        # l2.addWidget(self.widgetPlot)
        a99.set_margin(l2, 0)
        a = self.centralWidget = QWidget()
        a.setLayout(l2)
        a.setFont(a99.MONO_FONT)
        self.setCentralWidget(self.centralWidget)



        # # # Status bar
        # self.label_n = QLabel()
        # self.label_t = QLabel()
        # sb = self.statusBar()
        # sb.insertWidget(0, self.label_n, 0)
        # sb.insertWidget(1, self.label_t, 0)


        # # Final adjustments
        self.setGeometry(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT)
        a99.place_center(self)
        self.setWindowTitle("Runnables Manager")

        # # Non-visual stuff

        # ## Wiring of signals to trigger status update
        t = self.timerUpdate = QTimer()
        t.setInterval(1000)  # miliseconds
        signals = [t.timeout, self.rm.runnable_changed, self.rm.finished]
        self.changed_proxy = a99.SignalProxy(signals,
        delay=0, rateLimit=1, slot=self.__update, flag_connect=False)
        self.rm.runnable_added.connect(self.__populate, Qt.QueuedConnection)

        # ## Timer to flick the finish indicator
        t = self.timerFinished = QTimer()
        t.setInterval(200)  # miliseconds
        t.timeout.connect(self.on_timerFinished_timeout)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def showEvent(self, event):
        self.__populate()
        self.changed_proxy.connect_all()
        # self.rm.runnable_added.connect(self.on_tm_thread_added, Qt.QueuedConnection)
        self.timerUpdate.start()

    def closeEvent(self, event):
        if not self.rm.flag_finished:
            r = QMessageBox.question(self, "Close window", "Not finished yet. Are you sure?",
             QMessageBox.Yes|QMessageBox.No, QMessageBox.Yes)
            if r != QMessageBox.Yes:
                event.ignore()
                return
        self.changed_proxy.disconnect_all()
        # self.rm.runnable_added.disconnect(self.on_tm_thread_added)
        self.timerUpdate.stop()
        self.timerFinished.stop()

        if self.flag_close_mpl_plots_on_close:
            plt.close("all")

    def eventFilter(self, obj, event):
        if obj == self.tableWidget:
            return a99.check_return_space(event, self.on_tableWidget_cellDoubleClicked)
        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    # def on_tm_thread_changed(self):
    #     print "on_tm_thread_changed"
    #     self._update()

    # def on_tm_thread_added(self):
    #     self._populate()

    def on_pause(self):
        if self.pushButtonPause.isChecked():
            self.rm.pause()
        else:
            self.rm.resume()

    def on_status(self):
        self.widgetStatus.setVisible(self.pushButtonStatus.isChecked())

    def on_table(self):
        self.tableWidget.setVisible(self.pushButtonTable.isChecked())

    def on_exit(self):
        self.rm.cancel()

    def on_kill(self):
        self.rm.kill_runnables()

    def on_retry_failed(self):
        try:
            self.rm.retry_failed()
        except Exception as e:
            MSG = "Could not retry failed"
            a99.get_python_logger().exception(MSG)
            a99.show_error("%s: %s" % (MSG, str(e)))

    def on_collect_errors(self):
        # TODO **duplicate code** export a_XExplorer's one and use because the latter is more complete
        try:
            k = a99.ErrorCollector()
            k.collect_errors(".")
            w = f311.XHTML(self, k.get_html(), "Errors in '.' and subdirectories")
            w.show()
        except Exception as e:
            MSG = "Could not collect errors"
            a99.get_python_logger().exception(MSG)
            a99.show_error("%s: %s" % (MSG, str(e)))

    # def on_timer_timeout(self):
    #     print "on_timer_timeout"
    #     self._update()

    def on_timerFinished_timeout(self):
        self.__set_labelFinished_color(COLORS[self.__lfcs])
        self.__lfcs += 1
        if self.__lfcs == len(COLORS):
            self.__lfcs = 0

    def on_tableWidget_cellDoubleClicked(self, row=0, col=0):
        try:
            self.__explore_directory()
        except Exception as e:
            MSG = "Could explore directory"
            a99.get_python_logger().exception(MSG)
            a99.show_error("%s: %s" % (MSG, str(e)))


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Gear

    def __populate(self):
        """Clears and rebuilds table widget."""
        with self.__lock_table:
            a = self.tableWidget
            self.runnables = runnables = self.rm.get_runnables_copy()
            a.clear()
            self.__num_finished_shown = 0
            a.setAlternatingRowColors(True)
            a.setRowCount(len(runnables))
            a.setColumnCount(2)
            a.setHorizontalHeaderLabels(["Session directory", "Status"])
            for i, runnable in enumerate(runnables):
                title = runnable.sid.dir
                if title is None:
                    title = '...'
                item = QTableWidgetItem(title)
                a.setItem(i, 0, item)
                status = str(runnable.get_status())
                item = QTableWidgetItem("?" if status is None else str(status))
                a.setItem(i, 1, item)
            a.resizeColumnsToContents()
            a.setColumnWidth(1, STATUS_COLUMN_WIDTH)
            self.__update_status()

    def __update(self):
        """Updates second column of table widget."""
        # a99.get_python_logger().info("UPDATE UPDATE UPDATE UPDATE UPDATE ")
        t = time.time()
        self.__update_status()
        self.__update_table()
        # a99.get_python_logger().info("&&&&&&&&&&&&&&& time to update: %g" % (time.time()-t,))

    def __update_table(self):
        with self.__lock_table:
            a = self.tableWidget
            runnables = self.runnables
            nf = self.rm.num_finished  # grabs this before last table update,
                                       # so that it never skips a row update
            mt = self.rm.max_simultaneous
            for i in range(max(0, self.__num_finished_shown-mt+1),
                            min(len(runnables), self.rm.num_finished+mt)):
                runnable = runnables[i]

                item = a.item(i, 0)
                title = runnable.sid.dir
                if title is None:
                    title = "..."
                item.setText(title)

                item = a.item(i, 1)
                status = runnable.get_status()
                item.setText("?" if status is None else str(status))

            if nf != self.__num_finished_shown:
                a.setCurrentCell(self.rm.num_finished+self.rm.max_simultaneous-1, 0)
                self.__num_finished_shown = nf

    def __update_status(self):
        """Updates everything except the table widget."""
        flag_todo = not (self.rm.flag_finished or self.rm.flag_exited)
        self.pushButtonCancel.setEnabled(flag_todo)
        self.pushButtonPause.setEnabled(flag_todo)
        if not flag_todo and self.flag_close_message:
            s = "Please close window to continue"
            self.labelEnd.setText(s)
            self.labelEnd.setVisible(True)
            self.timerFinished.start()
        if self.pushButtonStatus.isChecked():
            self.labelStatusFinished.setText("%d/%d" %
             (self.rm.num_finished, self.rm.num_runnables))
            self.labelStatusFailed.setText(str(self.rm.num_failed))
            ella, tot, rema = self.rm.get_times()
            self.labelStatusElla.setText(a99.seconds2str(ella))
            self.labelStatusTPR.setText(a99.seconds2str(self.rm.time_per_runnable))
            self.labelStatusTotal.setText(a99.seconds2str(tot))
            self.labelStatusRema.setText(a99.seconds2str(rema))

            self.checkBox_paused.setChecked(self.rm.flag_paused)
            self.checkBox_cancelled.setChecked(self.rm.flag_cancelled)
            self.checkBox_finished.setChecked(self.rm.flag_finished)
            self.checkBox_failed.setChecked(self.rm.flag_failed)
            self.checkBox_exited.setChecked(self.rm.flag_exited)

    def __set_labelFinished_color(self, color):
        self.labelEnd.setStyleSheet("QLabel {color: "+color.name()+";}")

    def __explore_directory(self):
        runnable = self.runnables[self.tableWidget.currentRow()]
        dir_ = runnable.sid.dir
        if not self.__explorer_form:
            f = self.__explorer_form = f311.XExplorer(self, dir_)
            f.flag_close_mpl_plots_on_close = False
        else:
            self.__explorer_form.set_dir(dir_)
        self.__explorer_form.show()
        self.__explorer_form.raise_()
        self.__explorer_form.activateWindow()




__all__ = ["FileMolecules", "Molecule", "SetOfLines", "molconsts_to_molecule"]



import sys
import numpy as np
import a99
from f311 import DataFile
from .. import basic
import re

# TODO figure out state_from, state_to

@a99.froze_it
class SetOfLines(a99.AttrsPart):
    """
    Represents a "Set-Of-Lines" (vibrational transition with unique (vl, v2l))

    **Attention** Some properties return numpy arrays despite internally the data is stored as lists.
                  This was a bad idea but I won't change this anymore to avoid breaking client code

    Args:
        vl=None: upper vibrational state (v')
        v2l=None: lower vibrational state (v")
        qqv=None: Franck-Condon Factor
        ggv=None: TODO ?doc? (calculated from diatomic molecular constants)
        bbv=None: TODO ?doc? (calculated from diatomic molecular constants)
        ddv=None: TODO ?doc? (calculated from diatomic molecular constants)
        fact=None: TODO ?doc? (usually =1.)
        state_from=None: letter
        state_to=None: letter
    """

    attrs = ["vl", "v2l", "qqv", "ggv", "bbv", "ddv", "fact", "num_lines", "state_from", "state_to", "llzero", "llfin"]

    @property
    def llzero(self):
        """Minimum wavelength"""
        return min(self.lmbdam)

    @property
    def llfin(self):
        """Maximum wavelength"""
        return max(self.lmbdam)


    def __init__(self, vl=-1, v2l=-1, qqv=0, ggv=0, bbv=0, ddv=0, fact=0,
                 state_from="", state_to=""):
        a99.AttrsPart.__init__(self)

        self.vl = vl
        self.v2l = v2l
        self.qqv = qqv
        self.ggv = ggv
        self.bbv = bbv
        self.ddv = ddv
        self.fact = fact
        # # Transitional information (not stored in PFANT molecular lines file)
        # Example: 'A', 'X'
        self.state_from = state_from
        self.state_to = state_to

        # Vectors are keps as lists in order to .append_line() to work
        self.lmbdam = []
        self.sj = []
        self.jj = []
        self.branch = []

    @property
    def num_lines(self):
        return len(self)

    def __len__(self):
        return len(self.lmbdam)

    def __repr__(self):
        return "{}({}, {}, {}, {}, {}, {}, {})".format(self.__class__.__name__,
            self.vl, self.v2l, self.qqv, self.ggv, self.bbv, self.ddv, self.fact)

    def __iter__(self):
        """Creates MyDBRow objects to represent each molecular line"""
        fieldnames = ["lmbdam", "sj", "jj", "branch"]
        for t in zip(self.lmbdam, self.sj, self.jj, self.branch):
            obj = a99.MyDBRow()
            for fieldname, value in zip(fieldnames, t):
                obj[fieldname] = value
            yield obj

    def cut(self, lzero, lfin, jjmax=None):
        """Reduces the number of lines to only the ones whose lmbdam is inside [lzero, lfin]"""
        l, s, j, b = [], [], [], []
        for _l, _s, _j, _b in zip(self.lmbdam, self.sj, self.jj, self.branch):
            if lzero <= _l <= lfin and (jjmax is None or _j <= jjmax):
                l.append(_l)
                s.append(_s)
                j.append(_j)
                b.append(_b)
        self.lmbdam, self.sj, self.jj, self.branch = l, s, j, b

    def append_line(self, lmbdam, sj, jj, branch):
        self.lmbdam.append(lmbdam)
        self.sj.append(sj)
        self.jj.append(jj)
        self.branch.append(branch)

    def sort(self):
        """Sorts by wavelength **in-place**"""

        ii = np.argsort(self.lmbdam)
        self.lmbdam = [self.lmbdam[i] for i in ii]
        self.sj = [self.sj[i] for i in ii]
        self.jj = [self.jj[i] for i in ii]
        self.branch = [self.branch[i] for i in ii]


@a99.froze_it
class Molecule(a99.AttrsPart):
    attrs = ["description", "symbols", "fe", "do", "mm", "am", "bm", "ua", "ub",
             "te", "cro", "s", "nv", "num_lines", "llzero", "llfin"]

    @property
    def llzero(self):
        """Minimum wavelength"""
        return min([a.llzero for a in self.sol])

    @property
    def llfin(self):
        """Maximum wavelength"""
        return max([a.llfin for a in self.sol])

    @property
    def lmbdam(self):
        return np.hstack([np.array(x.lmbdam) for x in self.sol])

    @property
    def sj(self):
        return np.hstack([np.array(x.sj) for x in self.sol])

    @property
    def jj(self):
        return np.hstack([np.array(x.jj) for x in self.sol])

    @property
    def branch(self):
        return np.hstack([np.array(x.branch) for x in self.sol])

    @property
    def nv(self):
        return len(self.sol)

    @property
    def num_lines(self):
        ret = sum(map(len, self.sol))
        return ret

    @property
    def formula(self):
        return "".join([s[0]+s[1].lower() if len(s.strip()) == 2 else s.strip() for s in self.symbols])

    @property
    def qqv(self):
        return [x.qqv for x in self.sol]

    @property
    def ggv(self):
        return [x.qqv for x in self.sol]

    @property
    def bbv(self):
        return [x.qqv for x in self.sol]

    @property
    def ddv(self):
        return [x.qqv for x in self.sol]

    def __init__(self):
        a99.AttrsPart.__init__(self)

        # # "titulo" part
        # (20161206) titulo is now used semantically: it has 3 fields separated by a "#":
        # 'description # symbols # transitions'.
        #
        # For more information, refer to pfantlib.f90:read_molecules(),
        # look for where variable km_titulo is read from file
        #
        # "titulo" is loaded from file, but not used when saving the file
        self.titulo = ""
        self.description = ""
        self.symbols = None
        self.fe = 0.
        self.do = 0.
        self.mm = 0.
        self.am = 0.
        self.bm = 0.
        self.ua = 0.
        self.ub = 0.
        self.te = 0.
        self.cro = 0.
        self.s = 0.

        self.sol = []  # list of SetOfLines objects


    def __len__(self):
        """Returns number of set-of-lines."""
        return len(self.sol)

    def __iter__(self):
        return iter(self.sol)

    def __getitem__(self, *args):
        return self.sol.__getitem__(*args)


    def cut(self, lzero, lfin, jjmax=None):
        """Reduces the number of lines to only the ones whose lmbdam is inside [lzero, lfin]"""

        for i in reversed(list(range(len(self)))):
            self.sol[i].cut(lzero, lfin, jjmax)
            if len(self.sol[i]) == 0:
                del self.sol[i]


@a99.froze_it
class FileMolecules(DataFile):
    """
    PFANT Molecular Lines

    Rather than as read_molecules() in readers.f90, this class stores
    information for each molecule inside a Molecule object.
    """

    default_filename = "molecules.dat"
    attrs = ["titm", "num_lines"]
    editors = ["mled.py"]

    @property
    def llzero(self):
        """Minimum wavelength"""
        return min([a.llzero for a in self.molecules])

    @property
    def llfin(self):
        """Maximum wavelength"""
        return max([a.llfin for a in self.molecules])

    @property
    def num_lines(self):
        """Total number of spectral line, counting all molecules."""
        return sum([x.num_lines for x in self.molecules])

    @property
    def num_sols(self):
        """Total number of spectral line, counting all molecules."""
        return sum([len(x) for x in self.molecules])

    @property
    def lmbdam(self):
        return np.hstack([m.lmbdam for m in self.molecules])

    @property
    def sj(self):
        return np.hstack([m.sj for m in self.molecules])

    @property
    def jj(self):
        return np.hstack([m.jj for m in self.molecules])

    @property
    def branch(self):
        return np.hstack([m.branch for m in self.molecules])

    @property
    def qqv(self):
        return np.hstack([x.qqv for x in self.molecules])

    @property
    def ggv(self):
        return np.hstack([x.ggv for x in self.molecules])

    @property
    def bbv(self):
        return np.hstack([x.bbv for x in self.molecules])

    @property
    def ddv(self):
        return np.hstack([x.ddv for x in self.molecules])

    @property
    def description(self):
        return [m.description for m in self.molecules]

    @property
    def fe(self):
        return [m.fe for m in self.molecules]

    @property
    def do(self):
        return [m.do for m in self.molecules]

    @property
    def mm(self):
        return [m.mm for m in self.molecules]

    @property
    def am(self):
        return [m.am for m in self.molecules]

    @property
    def bm(self):
        return [m.bm for m in self.molecules]

    @property
    def ua(self):
        return [m.ua for m in self.molecules]

    @property
    def ub(self):
        return [m.ub for m in self.molecules]

    @property
    def te(self):
        return [m.te for m in self.molecules]

    @property
    def cro(self):
        return [m.cro for m in self.molecules]

    @property
    def s(self):
        return [m.s for m in self.molecules]

    def __init__(self):
        DataFile.__init__(self)

        # Array of Molecule objects
        self.molecules = []

        # Literal in second row of file, sortta used as a file title/description
        self.titm = ""

    def __len__(self):
        return len(self.molecules)

    def __iter__(self):
        return iter(self.molecules)

    def __getitem__(self, *args):
        return self.molecules.__getitem__(*args)

    def cut(self, lzero, lfin, jjmax=None):
        """Reduces the number of lines to only the ones whose lmbdam is inside [lzero, lfin]"""

        for i in reversed(list(range(len(self)))):
            m = self.molecules[i]
            m.cut(lzero, lfin, jjmax)
            if len(m) == 0:
                del self.molecules[i]

    def _do_load(self, filename):
        """Clears internal lists and loads from file."""

        with open(filename, "r") as h:
            r = 0 # counts rows of file
            try:
                number = int(h.readline())  # not used (see below)
                r += 1
                self.titm = a99.readline_strip(h)
                r += 1

                nv = a99.int_vector(h)  # number of transitions=sets-of-lines for each molecule
                r += 1
                # Uses length of nv vector to know how many molecules to read (ignores "number")
                num_mol = len(nv)

                for im in range(num_mol):
                    nvi = nv[im]

                    m = Molecule()
                    self.molecules.append(m)

                    m.titulo = a99.readline_strip(h)
                    a99.get_python_logger().debug('Reading %d%s molecule \'%s\'' % (im+1, a99.ordinal_suffix(im+1), m.titulo))

                    parts = [s.strip() for s in m.titulo.split("#")]
                    m.description = parts[0]
                    if len(parts) > 1:
                        # Found 'structure' in m.titulo
                        m.symbols = [basic.adjust_atomic_symbol(s) for s in
                                     [s.strip() for s in parts[1].split(" ") if len(s.strip()) > 0]]
                    else:
                        # Will try to guess molecule by m.titulo's contents
                        temp = basic.description_to_symbols(parts[0])
                        m.symbols = temp or []
                    transitions = []
                    if len(parts) > 2:
                        numbers = [int(float(x)) for x in re.findall('([0-9.]+)', parts[2])]
                        transitions = list(zip(numbers[0::2], numbers[1::2]))

                    r += 1
                    m.fe, m.do, m.mm, m.am, m.bm, m.ua, m.ub, m.te, m.cro = a99.float_vector(h)
                    r += 1

                    h.readline()  # Skips line which is blank in file
                                  # In readers.f90 the variables are ise, a0, a1, a2 a3, a4, als
                                  # but the pfant does not use them.
                    r += 1

                    m.s = float(h.readline())
                    r += 1

                    # These vectors must have nvi elements
                    s_v, r_inc = a99.multirow_str_vector(h, nvi, r)
                    r += r_inc
                    qqv = list(map(float, s_v))
                    s_v, r_inc = a99.multirow_str_vector(h, nvi, r)
                    r += r_inc
                    ggv = list(map(float, s_v))
                    s_v, r_inc = a99.multirow_str_vector(h, nvi, r)
                    r += r_inc
                    bbv = list(map(float, s_v))
                    s_v, r_inc = a99.multirow_str_vector(h, nvi, r)
                    r += r_inc
                    ddv = list(map(float, s_v))
                    s_v, r_inc = a99.multirow_str_vector(h, nvi, r)
                    r += r_inc
                    fact = list(map(float, s_v))
                    for name in ["qqv", "ggv", "bbv", "ddv", "fact"]:
                        v = eval(name)
                        if len(v) != nvi:
                            raise RuntimeError(
                                'Attribute %s of molecule #%d must be a vector with %d elements (has %d)' %
                                (name, im+1, nvi, len(v)))

                    # creates sets of lines and appends to molecule
                    for isol, (q, g, b, d, f) in enumerate(zip(qqv, ggv, bbv, ddv, fact)):
                        o = SetOfLines()
                        o.qqv = q
                        o.ggv = g
                        o.bbv = b
                        o.ddv = d
                        o.fact = f
                        if isol < len(transitions):
                            o.vl, o.v2l = transitions[isol]
                        m.sol.append(o)

                    # Now reads lines
                    sol_iter = iter(m.sol)  # iterator to change the current set-of-lines with the "numlin" flag
                    o = next(sol_iter)  # current set-of-lines
                    # o.lmbdam, o.sj, o.jj, o.branch = [], [], [], []
                    while True:
                        # Someone added "*" signs as a 6th column of some lines
                        # which was causing my reading to crash.
                        # Therefore I read the line and discard beyond the 5th column before
                        # converting to float
                        temp = a99.str_vector(h)
                        lmbdam = float(temp[0])
                        sj = float(temp[1])
                        jj = float(temp[2])
                        # Alphanumeric now iz = int(temp[3])
                        iz = temp[3]
                        numlin = int(temp[4])

                        r += 1

                        o.lmbdam.append(lmbdam)
                        o.sj.append(sj)
                        o.jj.append(jj)
                        o.branch.append(iz)

                        if numlin > 0:
                            if numlin == 9:
                                break
                            o = next(sol_iter)

                    a99.get_python_logger().info("Loading '{}': {}".format(filename, a99.format_progress(im+1, num_mol)))

                    if im+1 == num_mol:
                        break

                    # im += 1
            except Exception as e:
                raise type(e)(("Error around %d%s row of file '%s'" %
                    (r+1, a99.ordinal_suffix(r+1), filename))+": "+str(e)).with_traceback(sys.exc_info()[2])

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            a99.write_lf(h, str(len(self.molecules)))
            a99.write_lf(h, self.titm)
            a99.write_lf(h, " ".join([str(x.nv) for x in self.molecules]))
            for i_m, m in enumerate(self.molecules):
                # todo cleanup a99.get_python_logger().info("Saving '{}': molecule {}/{}".format(filename, i_m+1, len(self.molecules)))

                # # Assembles "titulo"
                # ## Transitions
                ltrans = []
                for sol in m:
                    if sol.vl is None or sol.v2l is None:
                        break
                    ltrans.append([sol.vl, sol.v2l])
                new_titulo = "{} # {} # {}".\
                    format(m.description,
                           " ".join([s.strip() for s in m.symbols]),
                           "|".join(["{:.0f},{:.0f}".format(*t) for t in ltrans]))

                # - mled change, incorporate shit

                a99.write_lf(h, new_titulo)
                a99.write_lf(h, (" ".join(["%.10g"]*9)) % (m.fe, m.do, m.mm, m.am,
                    m.bm, m.ua, m.ub, m.te, m.cro))
                a99.write_lf(h, "")
                a99.write_lf(h, str(m.s))
                a99.write_lf(h, " ".join(["{:g}".format(x.qqv) for x in m.sol]))
                a99.write_lf(h, " ".join(["{:.2f}".format(x.ggv) for x in m.sol]))
                a99.write_lf(h, " ".join(["{:.5f}".format(x.bbv) for x in m.sol]))
                a99.write_lf(h, " ".join(["{:.2f}".format(x.ddv) for x in m.sol]))
                a99.write_lf(h, " ".join(["{:g}".format(x.fact) for x in m.sol]))

                num_sol = len(m.sol)
                for i, s in enumerate(m.sol):
                    num_lines = len(s)  # number of lines for current set-of-lines
                    for j in range(num_lines):
                        numlin = 0 if j < num_lines-1 else 9 if i == num_sol-1 else 1
                        a99.write_lf(h, "%.10g %.10g %.10g %s %d" %
                                     (s.lmbdam[j], s.sj[j], s.jj[j], s.branch[j], numlin))


def molconsts_to_molecule(molconsts):
    """Assembles a Molecule object from record from a FileMolDB database

    Args:
        molconsts: a dict-like object combining field values from tables 'molecule', 'state',
                    'pfantmol', and 'system' from a FileMolDB database
    """
    molconsts["system_str"] = molconsts.get_system_str(basic.SS_PLAIN)

    mol = Molecule()
    mol.description = "{formula} [{system_str}]".format(**molconsts)
    mol.symbols = basic.description_to_symbols(molconsts["formula"])
    mol.fe = molconsts["fe"]
    mol.do = molconsts["do"]
    mol.mm = molconsts["am"] + molconsts["bm"]
    mol.am = molconsts["am"]
    mol.bm = molconsts["bm"]
    mol.ua = molconsts["ua"]
    mol.ub = molconsts["ub"]
    mol.te = molconsts["te"]
    mol.cro = molconsts.get_deltak()
    mol.s = molconsts.get_S2l()
    return mol


#!/usr/bin/env python

import argparse
import logging
import pyfant
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True


__doc__ = """PFANT Launcher -- Graphical Interface for Spectral Synthesis\n\nSingle and multi modes.\n\n
Multi mode
----------
"""+pyfant.gui._shared.DESCR_MULTI+\
"---------\n\n"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    args = parser.parse_args()

    app = a99.get_QApplication([])

    form0 = pyfant.XMulti()
    # this would start with "multi" tab selected form0.tabWidget.setCurrentIndex(3)

    rm = pyfant.RunnableManager()
    form1 = pyfant.XRunnableManager(form0, rm)
    form1.flag_close_mpl_plots_on_close = False
    form1.flag_close_message = False
    form0.set_manager_form(form1)
    # form1.show()
    # it is good to start the manager as late as possible, otherwise
    # the program will hang if, for example, the form fails to be created.
    rm.start()
    try:
        form0.show()
        app.exec_()
    finally:
        rm.exit()



__all__ = ["FileHmap", "HmapRow"]


import a99
from f311 import DataFile
import tabulate


@a99.froze_it
class HmapRow(a99.AttrsPart):
    """Same structure as pfantlib.f90::hmap_row type."""

    attrs = ["fn", "na", "nb", "clam", "kiex", "c1"]

    def __init__(self, fn=None, na=None, nb=None, clam=None, kiex=None, c1=None):
        a99.AttrsPart.__init__(self)
        self.fn = None
        self.na = None
        self.nb = None
        self.clam = None
        self.kiex = None
        self.c1 = None

    def __repr__(self):
        aa = ", ".join(["{}={}".format(key, self.__getattribute__(key)) for key in self.attrs])
        return "HMapRow({})".format(aa)


@a99.froze_it
class FileHmap(DataFile):
    """
    PFANT Hydrogen Lines Map

    Imitates the logic of reader_hmap.f90::read_hmap().

    Attributes match reader_hmap.f90::hmap_* (minus the "hmap_" prefix)
    """

    default_filename = "hmap.dat"
    attrs = ["rows"]

    def __init__(self):
        DataFile.__init__(self)

        # List of HmapRow objects
        self.rows = []

    def __str__(self):
        headers = HmapRow.attrs
        data = [[getattr(row, name) for name in headers] for row in self.rows]
        return tabulate.tabulate(data, headers)


    def __len__(self):
        return len(self.rows)

    def _do_load(self, filename):
        """Loads from file."""

        with open(filename, "r") as h:
            for i, line in enumerate(h):
                try:
                    line = line.strip()
                    if line.startswith("#") or len(line) == 0:
                        continue

                    r = HmapRow()
                    [r.fn, r.na, r.nb, r.clam, r.kiex, r.c1] = line.split()[:6]
                    [r.na, r.nb] = list(map(int, (r.na, r.nb)))
                    [r.clam, r.kiex, r.c1] = list(map(float, [r.clam, r.kiex, r.c1]))

                    self.rows.append(r)
                except:
                    # a99.get_python_logger().error("Error reading row #%d, file \"%s\"" % (i+1, filename))
                    raise


    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            a99.write_lf(h, "# filename / niv inf / niv sup / central lambda / kiex / c1")
            for r in self.rows:
                a99.write_lf(h, "%s %s %s %s %s %s" % (r.fn, r.na, r.nb, r.clam, r.kiex, r.c1))



from .convlog import *
from .conv import *
import pyfant

__all__ = ["ConvKurucz"]


class ConvKurucz(Conv):
    """Converts Kurucz molecular lines data to PFANT "sets of lines"

        Args:
            flag_spinl: Whether or not to use the spinl of the Kurucz line list
                        for branch determination (spin2l is always used). Effect:

                - True: possible branches are P1, P12, P21, P2 and repeat for Q, R
                        (12 possibilities in total)

                - False: possible branches are P1, P2 and repeat for Q, R
                         (6 possibilities in total)

            iso: (int or None) isotope. If specified as int, only that isotope will be filtered;
                 otherwise, all isotopes in file will be included. Isotope is
                 field KuruczMolLine.iso (see KuruczMolLine, FileKuruczMol).

                 **Note** old Kurucz file format does not have the isotope information, therefore,
                          iso must be None or the thing will not work

    """

    def __init__(self, flag_spinl=False, iso=None, *args, **kwargs):
        Conv.__init__(self, *args, **kwargs)
        self.flag_spinl = flag_spinl
        self.iso = iso

    def _make_sols(self, lines):

        log = self.log
        sols = self.sols

        def append_error(msg):
            log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))

        if not isinstance(lines, pyfant.FileKuruczMoleculeBase):
            raise TypeError("Invalid type for argument 'fileobj': {}".format(type(lines).__name__))
        if self.mode == ConvMode.F:
            if not isinstance(lines, pyfant.FileKuruczMolecule):
                raise ValueError("Old-format file does not contain oscillator strength")

            deltak = self.molconsts.get_deltak()
            S2l = self.molconsts.get_S2l()
        elif self.mode == ConvMode.HLF:
            pass
        else:
            raise ValueError(f"mode=={self.mode} not implemented")


        lines = lines.lines
        log.n = len(lines)
        if log.n == 0:
            raise RuntimeError("Zero lines found")


        STATEL = self.molconsts["from_label"]
        STATE2L = self.molconsts["to_label"]

        mtools = self.kovacs_toolbox()

        for i, line in enumerate(lines):
            if self.iso and line.iso != self.iso:
                log.skip_reasons["Isotope {}".format(line.iso)] += 1
                continue

            if line.statel != STATEL or line.state2l != STATE2L:
                log.skip_reasons["Transition {}-{}".format(line.statel, line.state2l)] += 1
                continue

            branch = mtools.quanta_to_branch(line.Jl, line.J2l,
                spinl=(None if not self.flag_spinl else line.spinl), spin2l=line.spin2l)

            # This was only a test, but filters like these may be useful if line.spin2l != line.spinl:
            #     log.skip_reasons["Different spin"] += 1
            #     continue

            try:
                if self.mode == ConvMode.HLF:
                    sj = self.get_hlf(line.vl, line.v2l, line.J2l, branch)
                else:
                    k = 1/((2*S2l+1)*(2-deltak)*(2*line.J2l+1))
                    sj = k*10**line.loggf

                self.append_line(line, sj, branch)

            except Exception as e:
                reason = a99.str_exc(e)
                log.skip_reasons[reason] += 1
                msg = "#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), reason)
                log.errors.append(msg)
                if not self.flag_quiet:
                    a99.get_python_logger().exception(msg)
                continue



__all__ = ["FileToH"]

import math
import numpy as np
import fortranformat as ff
import a99
from f311 import DataFile


class FileToH(DataFile):
    """
    PFANT Hydrogen Line Profile

    Imitates the logic of reader_filetoh.f90::read_filetoh(). Difference: pfant
    reads second row as single character-variable "ttt", but hydro2 saves 5 values
    in second row. I am reading these values.

    Attributes match reader_filetoh.f90:filetoh_r_* (minus the "filetoh_r_" prefix)
    """
    default_filename = "thalpha"

    attrs = ["titre", "ntot", "zut1", "zut2", "zut3", "th", "lambdh", "jmax"]

    def __init__(self):
        DataFile.__init__(self)
        self.titre = None

        # second row
        self.ntot = None
        self.zut1 = None
        self.zut2 = None
        self.zut3 = None

        # will be NumPy matrix [jmax]x[modeles_ntot]
        self.th = None
        self.lambdh = None
        self.jmax = None


    def _do_load(self, filename):
        with open(filename, "r") as h:
            #-- row 1
            self.titre = h.readline().strip('\n')

            #-- row 2
            fr = ff.FortranRecordReader('(I6,2F6.1,2X,2F8.4)')
            vars = fr.read(h.readline())
            [self.ntot, self.zut1, _, self.zut2, self.zut3] = vars

            #-- row 3
            self.jmax = int(h.readline())


            #-- lambda vector
            #-- Reads sequence of jmax float numbers of 14 characters each disposed in
            #-- rows of 5 columns or less
            # function to calculate number of 5-column rows needed to store a sequence
            num_rows = lambda x: int(math.ceil(float(x)/5))
            FLOAT_SIZE = 14  # size of stored float value in characters
            nr = num_rows(self.jmax)
            v = []
            for i in range(nr):
                v.extend([float(x) for x in a99.chunk_string(a99.readline_strip(h), FLOAT_SIZE)])
            self.lambdh = np.array(v)

            if not (len(self.lambdh) == self.jmax):
                raise RuntimeError("len(lambdh)=%d should be %d" % (len(self.lambdh), self.jmax))

            #-- (lambda) x (atmospheric layer) matrix
            v = []  # Will accumulate values for future reshape
            FLOAT_SIZE = 12  # size of stored float value in characters
            for s in h.readlines():
                s = s.strip('\n')
                v.extend([float(x) for x in a99.chunk_string(s, FLOAT_SIZE)])
            if len(v)/self.jmax != self.ntot:
                raise RuntimeError("Should have found %d values for th matrix (found %d)" %
                                           (self.jmax*self.ntot, len(v)))

            # Note that matrix is filled horizontally, i.e., first row, then second row, ...
            # whereas it was dumped in the file by column; hence we have to transpose it
            self.th = np.reshape(v, (self.ntot, self.jmax)).T

    def _do_save_as(self, filename):
        """Saves to file."""
        raise NotImplementedError("Maybe tomorrow")

    def init_default(self):
        raise RuntimeError("Not applicable")




"""Widget to edit a FileMain object."""

__all__ = ["WFileMolConsts"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from .a_WMolecularConstants import *
import pyfant


class WFileMolConsts(a99.WEditor):
    """FileMolConsts editor widget."""

    @property
    def fcfs(self):
        if self.w_molconsts is None:
            return None
        return self.w_molconsts.fcfs

    def __init__(self, *args, **kwargs):
        a99.WEditor.__init__(self, *args, **kwargs)

        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False

        # # Central layout
        l = self.layout_editor

        l.addLayout(self.keep_ref(self._get_layout_fn_moldb()))

        w = self.w_molconsts = WMolecularConstants(self.parent_form)
        w.changed.connect(self._on_w_molconsts_changed)
        l.addWidget(w)

        # ## Adds vertically expanding spacer at bottom
        l.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        self.setEnabled(False)  # disabled until load() is called
        self.flag_process_changes = True

    def _get_layout_fn_moldb(self):
        help = "This is a file such as '{}', containing a list of diatomic molecules, their electronic systems, and constants for each electronic state.\n\n" \
               "To create a new such file, or make a copy of PyFANT default one in your local directory, run `moldbed.py` first.".format(pyfant.FileMolDB.default_filename)
        l = QHBoxLayout()
        a = self.keep_ref(QLabel("Molecular constants database file"))
        w = self.w_fn_moldb = a99.WSelectFile(self.parent_form)
        w.setToolTip(help)
        a.setBuddy(w)
        w.changed.connect(self._on_fn_moldb_changed)
        l.addWidget(a)
        l.addWidget(w)
        l.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        return l

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    # def set_moldb(self, fobj):
    #     self.w_molconsts.set_moldb(fobj)

    def _do_load(self, fobj):
        assert isinstance(fobj, pyfant.FileMolConsts)
        self._f = fobj
        self.w_fn_moldb.value = fobj.fn_moldb
        self._do_load_moldb()
        self.w_molconsts.molconsts = fobj.molconsts
        # The only thing that can go wrong (GUI bein invalid-wise) is the moldb filename missing
        self._flag_valid = self.w_fn_moldb.flag_valid

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""
        self.w_molconsts.setFocus()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def _on_w_molconsts_changed(self):
        if self.flag_process_changes:
            self._flag_valid = self.w_molconsts.flag_valid
            self.changed.emit()

    def _on_fn_moldb_changed(self):
        self._f.fn_moldb = self.w_fn_moldb.value
        self._do_load_moldb()
        self.changed.emit()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internals


    def _do_load_moldb(self):
        """Loads FileMolDB and updates part of GUI accordingly.

        This is an intermediate GUI updating stage."""
        moldb, flag_valid = pyfant.FileMolDB(), False
        fn_moldb = self.w_fn_moldb.value
        try:
            moldb.load(fn_moldb)
            flag_valid = True
        except Exception as e:
            self.add_log_error("Error loading file '{}'".format(fn_moldb), True, e)
        self.w_fn_moldb.flag_valid = flag_valid
        self.w_molconsts.set_moldb(moldb if flag_valid else None)


import os
import glob
import platform
import a99


__all__ = [
    "get_pfant_path", "get_pfant_data_path", "get_pfant_data_subdirs", "get_pfant_star_subdirs",
    "get_fortrans",
    ]


def get_pfant_path(*args):
    """
    Returns absolute path to the "PFANT" directory.

    PFANT binaries must be in PATH,
    otherwise it will not work

    Arguments passed will be incorporated into path
    """

    path_to_exe = a99.which("pfant")
    if path_to_exe is None:
        raise RuntimeError("Cannot find 'pfant' executable")

    p, _ = os.path.split(path_to_exe)

    return os.path.abspath(os.path.join(p, "..", "..", *args))


def get_pfant_data_path(*args):
    """Returns absolute path to PFANT/data"""
    return get_pfant_path("data", *args)


def get_pfant_data_subdirs():
    """returns a list containing all subdirectories of PFANT/data (their names only, not full path)."""
    dd = glob.glob(os.path.join(get_pfant_data_path(), "*"))
    ret = []
    for d in dd:
        if os.path.isdir(d):
            ret.append(os.path.basename(d))
    return ret


def get_pfant_star_subdirs():
    """Returns only subdirectories of PFANT/data that contain file main.dat."""
    dd = glob.glob(os.path.join(get_pfant_data_path(), "*"))
    ret = []
    for d in dd:
        if os.path.isdir(d) and os.path.isfile(os.path.join(d, 'main.dat')):
            ret.append(os.path.basename(d))
    return ret


def get_fortrans(max_len=None):
    """
    Generates listing of files in the Fortran bin directory

    Arguments
      max_len: (optional) if passed, will use it, otherwise will be the
                 maximum number of characters among all Fortran executable names.

    Returns: list of a99.ExeInfo
    """

    ret = []

    flag_bindir = True
    try:
        bindir = os.path.join(get_pfant_path(), "fortran", "bin")
    except:
        flag_bindir = False

    ihpn = ["innewmarcs", "hydro2", "pfant", "nulbad"]
    if max_len is None:
        max_len = max(len(x) for x in ihpn)

    if platform.system() == "Windows":
        def has_it(x):
            return os.path.isfile(os.path.join(bindir, x+".exe"))
    else:
        def has_it(x):
            return os.path.isfile(os.path.join(bindir, x))
    for name in ihpn:
        status = "not found"
        if flag_bindir and has_it(name):
            status = "found"

        ret.append(a99.ExeInfo(name, status, flag_gui=None))
        # piece = name + " " + ("." * (max_len-len(name)))
        # ret.append(("%-"+str(max_len)+"s %s") % (piece, status))
    return ret





from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from a99 import WDBRegistry
import a99
import pyfant

__all__ = ["WDBState"]


# Field names to leave out of table widget
_FIELDNAMES_OUT = ("id_molecule",)


class WDBState(WDBRegistry):
    """Registry for table 'state'"""

    def __init__(self, *args):
        WDBRegistry.__init__(self, *args)

        self._id_molecule = None

        # Adds button to download NIST data
        action = self.action_download = QAction(a99.get_icon("alien"), "&Download from NIST", self)
        action.triggered.connect(self._on_download)
        self.toolbar.addAction(action)


    def set_id_molecule(self, id_):
        """Sets molecule id and re-populates table"""
        self._id_molecule = id_
        self._populate()
        self._move_to_first()

    def find_formula(self, formula):
        """Moves to row where formula is (if found, otherwise does nothing)"""
        for i, row in enumerate(self._data):
            if row["formula"] == formula:
                self.tableWidget.setCurrentCell(i, 0)
                break

    def _on_download(self):
        if self._id_molecule is None:
            return

        formula = self._f.get_conn().execute("select formula from molecule where id = ?",
                                             (self._id_molecule,)).fetchone()["formula"]

        n = len(self._data)
        beware = "" if n == 0 else \
            "\n\n**Attention**: {} existing state{} for this molecule will be deleted!".\
                format(n, "" if n == 1 else "s")

        msg = "Molecular constants for '{}' will be downloaded " \
              "from NIST chemistry web book{}\n\nDo you wish to continue?".format(formula, beware)

        r = QMessageBox.question(self, "Download NIST data", msg,
                                 QMessageBox.Yes | QMessageBox.No,
                                 QMessageBox.Yes if n == 0 else QMessageBox.No)

        if r == QMessageBox.Yes:
            try:
                nist_data, _, _ = pyfant.get_nist_webbook_constants(formula, flag_parse_state=True)
            except Exception as e:
                self.add_log_error("Error getting NIST data (see log for more info)", True, e)
            else:
                pyfant.insert_states_from_nist(self._f, self._id_molecule, nist_data, flag_replace=True)
                self._populate()
                n = len(self._data)
                self.add_log("{} state{} successfully downloaded from NIST".
                             format(n, "" if n == 1 else "s"), True)

    def _populate(self, restore_mode=None):
        """
        Populates table widget

        Args:
            restore_mode: how to reposition cursor:
                - "index": tries to restore same row index
                - anything else: does not re-position
        """

        if self._f is None:
            return

        self._flag_populating = True
        try:
            curr_idx = self.tableWidget.currentRow()

            t = self.tableWidget
            rows = a99.cursor_to_rows(self._f.query_state(id_molecule=self._id_molecule))
            ti = self._f.get_table_info("state")
            fieldnames = [name for name in ti if name not in _FIELDNAMES_OUT]
            # unfortunately QTableWidget is not prepared to show HTML col_names = [row["caption"] or row["name"] for row in ti if not row["name"] in FIELDNAMES_OUT]
            col_names = fieldnames
            nr, nc = len(rows), len(fieldnames)

            a99.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(col_names)

            if nr > 0:
                for i, row in enumerate(rows):
                    for j, fieldname in enumerate(fieldnames):
                        cell = row[fieldname]
                        item = QTableWidgetItem("" if cell is None else str(cell))
                        t.setItem(i, j, item)
            t.resizeColumnsToContents()
            self._data = rows

            if restore_mode == "index":
                if -1 < curr_idx < t.rowCount():
                    t.setCurrentCell(curr_idx, 0)
        finally:
            self._flag_populating = False
            self._wanna_emit_id_changed()

    def _get_edit_params(self):
        """Returns a Parameters object containing information about the fields that may be edited"""
        ti = self._f.get_table_info("state")
        params = a99.table_info_to_parameters(ti)
        params = [p for p in params if not p.name.startswith("id")]
        return params

    def _get_edit_field_names(self):
        """Returns a list with the names of the fields that may be edited"""
        params = self._get_edit_params()
        ret = [p.name for p in params]
        return ret

    def _do_on_insert(self):
        params = self._get_edit_params()
        form = a99.XParametersEditor(specs=params, title="Insert Molecular State")
        r = form.exec_()
        if r == QDialog.Accepted:
            kwargs = form.get_kwargs()
            conn = self._f.get_conn()
            s = "insert into state (id_molecule, {}) values ({}, {})".\
                format(", ".join([p.name for p in params]),
                       self._id_molecule,
                       ", ".join(["?"]*len(params)))
            cursor = conn.execute(s, list(kwargs.values()))
            id_ = cursor.lastrowid
            conn.commit()
            self._populate(False)
            self._find_id(id_)
            return True

    def _do_on_edit(self):
        r, form = a99.show_edit_form(self.row, self._get_edit_field_names(), "Edit Molecular State")
        if r == QDialog.Accepted:
            kwargs = form.get_kwargs()
            id_ = self.row["id"]
            s = "update state set {} where id = {}".format(
                ", ".join(["{} = '{}'".format(a, b) for a, b in kwargs.items()]), id_)
            conn = self._f.get_conn()
            conn.execute(s)
            conn.commit()
            self._populate()
            self._find_id(id_)
            return True

    def _do_on_delete(self):
        r = QMessageBox.question(None, "Delete Molecular State",
                                 "Are you sure?",
                                 QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if r == QMessageBox.Yes:
            conn = self._f.get_conn()
            id_ = self.row["id"]
            conn.execute("delete from state where id = ?", [id_])
            conn.commit()
            self._populate("index")
            return True



#!/usr/bin/env python
"""
Creates symbolic links to PFANT data files as an alternative to copying these (sometimes large) files into local directory

A star is specified by three data files whose typical names are:
main.dat, abonds.dat, and dissoc.dat .

The other data files (atomic/molecular lines, partition function, etc.)
are star-independent, and this script is a proposed solution to keep you from
copying these files for every new case.

How it works: link.py will look inside a given directory and create
symbolic links to files *.dat and *.mod.

The following files will be skipped:
  - main files, e.g. "main.dat"
  - dissoc files, e.g., "dissoc.dat"
  - abonds files, e.g., "abonds.dat"
  - .mod files with a single model inside, e.g., "modeles.mod"
  - hydrogen lines files, e.g., "thalpha", "thbeta"

This script works in two different modes:

a) default mode: looks for files in a subdirectory of PFANT/data
   > link.py common
   (will create links to filess inside PFANT/data/common)

b) "-l" option: lists subdirectories of PFANT/data

c) "-p" option: looks for files in a directory specified.
   Examples:
   > link.py -p /home/user/pfant-common-data
   > link.py -p ../../pfant-common-data

Note: in Windows, this script must be run as administrator.
"""
import argparse
import logging
import sys
import pyfant
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=a99.SmartFormatter
    )
    parser.add_argument('-l', '--list', action='store_true',
      help='lists subdirectories of '+pyfant.get_pfant_data_path())
    parser.add_argument('-p', '--path', action='store_true',
      help='system path mode')
    parser.add_argument('-y', '--yes', action="store_true",
                        help="Automatically answers 'yes' to eventual question")
    parser.add_argument('directory', type=str, nargs="?", default="common",
     help='name of directory (either a subdirectory of PFANT/data or the path '
          'to a valid system directory (see modes of operation)')

    args = parser.parse_args()

    # if len(sys.argv) == 1:
    #     args.list = True  # makes "-l" the default behaviour

    if (not args.directory or len(args.directory) ==  0) and not args.list:
        print("Directory name is required, except if '-l' option specified.")
        parser.print_usage()
        sys.exit()

    # "-l" mode
    if args.list:
        print("\n".join(a99.format_h1("Subdirectories of '%s'" % pyfant.get_pfant_data_path())))
        for dirname in pyfant.get_pfant_data_subdirs():
            print(dirname)
        sys.exit()

    if args.path:
        dir_ = args.directory
    else:
        dir_ = pyfant.get_pfant_path('data', args.directory)

    flag_ask = not args.path and not args.yes

    if flag_ask:
        while True:
            ans = input("Create links to PFANT data files in '%s' (Y/n)? " %
                            dir_).upper()
            if ans in ("N", "NO"):
                sys.exit()
            if ans in ("Y", "YES", ""):
                break

    pyfant.link_to_data(dir_)




"""Widget to edit a FileMain object."""

__all__ = ["WFileMain"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from ._shared import *
import a99
import pyfant


class WFileMain(QWidget):
    """
    FileMain editor widget.

    Args:
      parent=None
    """

    # Emitted whenever any value changes
    changed = pyqtSignal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        # Whether all the values in the fields are valid or not
        self.flag_valid = False
        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False
        self.f = None # FileMain object


        # # Central layout
        la = self.layout_main = QVBoxLayout()
        a99.set_margin(la, 0)
        self.setLayout(la)

        # ## Splitter with scroll area and descripton+error area
        sp = self.splitter = QSplitter(Qt.Vertical)
        la.addWidget(sp)


        # ### Widget + grid layout to be first in splitter
        sa = self.c33441 = QScrollArea()
        sp.addWidget(sa)
        sa.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        sa.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        # Widget that will be handled by the scrollable area
        w = self.scrollWidget = QWidget()
        sa.setWidget(self.scrollWidget)
        sa.setWidgetResizable(True)
#        la.addWidget(w)


        lw = QVBoxLayout()
        w.setLayout(lw)
        # Form layout
        lg = self.formLayout = QGridLayout()
        lw.addLayout(lg)
        a99.set_margin(lg, 0)
        lg.setVerticalSpacing(4)
        lg.setHorizontalSpacing(5)
        lw.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))



        # field map: [(label widget, edit widget, field name, short description,
        #              field name color, long description), ...]
        map = self._map = []

        x = self.label_titrav = QLabel()
        y = self.lineEdit_titrav = QLineEdit()
        # y.editingFinished.connect(self._on_editing_finished)
        y.textEdited.connect(self.on_edited)
        y.installEventFilter(self)
        x.setBuddy(y)
        map.append((x, y, "t&itrav", "star name", COLOR_STAR,
         "Name of the star"))

        x = self.label_teff = QLabel()
        y = self.lineEdit_teff = QLineEdit()
        y.textEdited.connect(self.on_edited)
        y.installEventFilter(self)
        y.setValidator(QDoubleValidator(0, 1e10, 0))
        x.setBuddy(y)
        map.append((x, y, "&teff", "effective temperature", COLOR_STAR,
         "Sun: 5777"))

        x = self.label_glog = QLabel()
        y = self.lineEdit_glog = QLineEdit()
        y.textEdited.connect(self.on_edited)
        y.installEventFilter(self)
        y.setValidator(QDoubleValidator(0, 1e10, 5))
        x.setBuddy(y)
        map.append((x, y, "&glog", "gravity", COLOR_STAR,
         "Sun: 4.44"))

        x = self.label_asalog = QLabel()
        y = self.lineEdit_asalog = QLineEdit()
        y.textEdited.connect(self.on_edited)
        y.installEventFilter(self)
        y.setValidator(QDoubleValidator(-10, 10, 5))
        x.setBuddy(y)
        map.append((x, y, "&asalog", "metallicity", COLOR_STAR,
         "Sun: 0"))

        x = self.label_vvt = QLabel()
        y = self.lineEdit_vvt = QLineEdit()
        y.textEdited.connect(self.on_edited)
        y.installEventFilter(self)
        y.setValidator(QDoubleValidator(0, 900, 2))
        x.setBuddy(y)
        map.append((x, y, "&vvt", "velocity of microturbulence", COLOR_STAR,
         "Sun: 0.9"))

        x = self.label_nhe = QLabel()
        y = self.lineEdit_nhe = QLineEdit()
        y.textEdited.connect(self.on_edited)
        y.installEventFilter(self)
        y.setValidator(QDoubleValidator(0, 10, 5))
        x.setBuddy(y)
        map.append((x, y, "&nhe", "abundance of Helium", COLOR_STAR,
         "Sun: 0.1"))

        x = self.label_ptdisk = QLabel()
        y = self.checkBox_ptdisk = QCheckBox()
        y.setTristate(False)
        y.installEventFilter(self)
        y.stateChanged.connect(self.on_edited)
        x.setBuddy(y)
        map.append((x, y, "pt&disk", "point of disk?", COLOR_CONFIG,
         DESCR_PTDISK))

        x = self.label_mu = QLabel()
        y = self.lineEdit_mu = QLineEdit()
        y.installEventFilter(self)
        y.textEdited.connect(self.on_edited)
        y.setValidator(QDoubleValidator(-1, 1, 5))
        x.setBuddy(y)
        map.append((x, y, "&mu", "cosine of angle", COLOR_CONFIG,
         "This is the cosine of the angle formed by center of "
         "the star disk, the point of observation, and the Earth as vertex. "
         "<br><br>This value will be used only if "+a99.enc_name("ptdisk", COLOR_CONFIG)+
         " is True.<br>\n<pre>\n"
         "             point of observation   \n"
         "            /                       \n"
         "           /                        \n"
         "          /                         \n"
         "         /                          \n"
         "        /                           \n"
         "       / ) acos(mu)                 \n"
         "Earth /------------------- center of star disk\n</pre>"))

        x = self.label_flprefix = QLabel()
        y = self.lineEdit_flprefix = QLineEdit()
        # y.editingFinished.connect(self._on_editing_finished)
        y.textEdited.connect(self.on_edited)
        y.installEventFilter(self)
        x.setBuddy(y)
        map.append((x, y, "flprefi&x", "prefix of filename", COLOR_CONFIG,
         "pfant will create three output files:<ul>"
         "<li>"+a99.enc_name("flprefix", COLOR_CONFIG)+".cont (continuum),"
         "<li>"+a99.enc_name("flprefix", COLOR_CONFIG)+".norm (normalized spectrum), and"
         "<li>"+a99.enc_name("flprefix", COLOR_CONFIG)+".spec (continuum*normalized)</ul>"))

        x = self.label_pas = QLabel()
        y = self.lineEdit_pas = QLineEdit()
        y.installEventFilter(self)
        y.textEdited.connect(self.on_edited)
        y.setValidator(QDoubleValidator(0, 10, 5))
        x.setBuddy(y)
        map.append((x, y, "&pas", "calculation step (&Aring;)", COLOR_CONFIG,
         "The synthetic spectrum will have points "+a99.enc_name("pas", COLOR_CONFIG)+
         " &Aring; distant from "
         "each other.<br><br>Use this to specify the resolution of the synthetic spectrum."))

        x = self.label_llzero = QLabel()
        y = self.lineEdit_llzero = QLineEdit()
        y.installEventFilter(self)
        y.textEdited.connect(self.on_edited)
        y.setValidator(QDoubleValidator(0, 10, 5))
        x.setBuddy(y)
        map.append((x, y, "ll&zero", "lower boundary of synthesis interval (&Aring;)",
                   COLOR_CONFIG, LLZERO_LLFIN))

        x = self.label_llfin = QLabel()
        y = self.lineEdit_llfin = QLineEdit()
        y.installEventFilter(self)
        y.textEdited.connect(self.on_edited)
        y.setValidator(QDoubleValidator(0, 10, 5))
        x.setBuddy(y)
        map.append((x, y, "&llfin", "upper boundary of synthesis interval (&Aring;)",
                   COLOR_CONFIG, LLZERO_LLFIN))

        x = self.label_aint = QLabel()
        y = self.lineEdit_aint = QLineEdit()
        y.installEventFilter(self)
        y.textEdited.connect(self.on_edited)
        y.setValidator(QDoubleValidator(0, 10, 5))
        x.setBuddy(y)
        map.append((x, y, "&aint", "length of sub-interval (&Aring;)", COLOR_CONFIG,
        "This is length of each calculation sub-interval "
        "(the calculation interval ["+a99.enc_name("llzero", COLOR_CONFIG)+", "+
        a99.enc_name("llfin", COLOR_CONFIG)+"] is split in sub-intervals of roughly "+
        a99.enc_name("aint", COLOR_CONFIG)+" &Aring;)."
        "<br><br>Note: "+a99.enc_name("aint", COLOR_CONFIG)+" must be a multiple of "+
        a99.enc_name("pas", COLOR_CONFIG)+"."+LLZERO_LLFIN))

        x = self.label_fwhm = QLabel()
        y = self.lineEdit_fwhm = QLineEdit()
        y.installEventFilter(self)
        y.textEdited.connect(self.on_edited)
        y.setValidator(QDoubleValidator(0, 10, 5))
        x.setBuddy(y)
        map.append((x, y, "f&whm", "convolution full-width-half-maximum", COLOR_CONFIG,
         "This parameter specifies the full-width-half-maximum "
         "of a Gaussian curve to convolve the synthetic spectrum with. <br><br>It is "
         "used by <em>nulbad</em> (Fortran code that calculates such convolution)."))


        for i, (label, edit, name, short_descr, color, long_descr) in enumerate(map):
            # label.setStyleSheet("QLabel {text-align: right}")
            assert isinstance(label, QLabel)
            label.setText(a99.enc_name_descr(name, short_descr, color))
            label.setAlignment(Qt.AlignRight)
            lg.addWidget(label, i, 0)
            lg.addWidget(edit, i, 1)
            label.setToolTip(long_descr)
            edit.setToolTip(long_descr)


        # ### Second widget of splitter
        # layout containing description area and a error label
        wlu = QWidget()
        lu = QVBoxLayout(wlu)
        a99.set_margin(lu, 0)
        lu.setSpacing(1)
        x = self.c23862 = QLabel("<b>Help</b>")
        lu.addWidget(x)
        x = self.textEditDescr = QTextEdit(self)
        x.setReadOnly(True)
        x.setStyleSheet("QTextEdit {color: %s}" % a99.COLOR_DESCR)
        lu.addWidget(x)
        x = self.c23862 = QLabel("<b>Errors</b>")
        lu.addWidget(x)
        x = self.labelError = QLabel(self)
        x.setStyleSheet("QLabel {color: %s}" % a99.COLOR_ERROR)
        lu.addWidget(self.labelError)
        sp.addWidget(wlu)


        self.setEnabled(False)  # disabled until load() is called
        a99.style_checkboxes(self)
        self.flag_process_changes = True


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def load(self, x):
        assert isinstance(x, pyfant.FileMain)
        self.f = x
        self.__update_guiile_main()
        # this is called to perform file validation upon loading
        self.__update_file_main()
        self.setEnabled(True)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""
        self.lineEdit_titrav.setFocus()

    def eventFilter(self, obj_focused, event):
        if event.type() == QEvent.FocusIn:
            text = ""
            for label, obj, name, short_descr, color, long_descr in self._map:
                if obj_focused == obj:
                    text = "%s<br><br>%s" % \
                           (a99.enc_name(name.replace("&", ""), color), long_descr)
                    break
            self.__set_descr_text(text)
        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_edited(self):
        if self.flag_process_changes:
            self.__update_file_main()
            self.changed.emit()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear

    def __update_guiile_main(self):
        self.flag_process_changes = False
        try:
            o = self.f
            self.lineEdit_titrav.setText(o.titrav)
            self.lineEdit_teff.setText(str(o.teff))
            self.lineEdit_glog.setText(str(o.glog))
            self.lineEdit_asalog.setText(str(o.asalog))
            self.lineEdit_vvt.setText(str(o.vvt[0]))
            self.lineEdit_nhe.setText(str(o.nhe))
            self.checkBox_ptdisk.setChecked(o.ptdisk)
            self.lineEdit_mu.setText(str(o.mu))
            self.lineEdit_flprefix.setText(o.flprefix)
            self.lineEdit_pas.setText(str(o.pas))
            self.lineEdit_llzero.setText(str(o.llzero))
            self.lineEdit_llfin.setText(str(o.llfin))
            self.lineEdit_aint.setText(str(o.aint))
            self.lineEdit_fwhm.setText(str(o.fwhm))
        finally:
            self.flag_process_changes = True

    def __set_error_text(self, x):
        """Sets text of labelError."""
        self.labelError.setText(x)

    def __set_descr_text(self, x):
        """Sets text of labelDescr."""
        self.textEditDescr.setText(x)

    def __update_file_main(self):
        o = self.f
        emsg, flag_error = "", False
        ss = ""
        try:
            ss = "titrav"
            o.titrav = str(self.lineEdit_titrav.text())
            ss = "teff"
            o.teff = float(self.lineEdit_teff.text())
            ss = "glog"
            o.glog = float(self.lineEdit_glog.text())
            ss = "asalog"
            o.asalog = float(self.lineEdit_asalog.text())
            o.afstar = o.asalog  # makes afstar match asalog
            ss = "vvt"
            temp = float(self.lineEdit_vvt.text())
            if temp > 900:
                raise RuntimeError("Not prepared for depth-variable velocity "
                 "of microturbulence (maximum allowed: 900)")
            o.vvt = [temp]
            ss = "nhe"
            o.nhe = float(self.lineEdit_nhe.text())
            ss = "ptdisk"
            o.ptdisk = self.checkBox_ptdisk.isChecked()
            ss = "mu"
            o.mu = float(self.lineEdit_mu.text())
            ss = "flprefix"
            o.flprefix = str(self.lineEdit_flprefix.text())
            ss = "pas"
            o.pas = float(self.lineEdit_pas.text())
            ss = "llzero"
            o.llzero = float(self.lineEdit_llzero.text())
            ss = "llfin"
            o.llfin = float(self.lineEdit_llfin.text())
            ss = "aint"
            o.aint = float(self.lineEdit_aint.text())
            ss = "fwhm"
            o.fwhm = float(self.lineEdit_fwhm.text())

            # Some error checks
            ss = ""
            if o.llzero >= o.llfin:
                raise RuntimeError("llzero must be lower than llfin!")
            if not (-1 <= o.mu <= 1):
                raise RuntimeError("mu must be between -1 and 1!")
        except Exception as E:
            flag_error = True
            if ss:
                emsg = "Field \"%s\": %s" % (ss, str(E))
            else:
                emsg = str(E)
            # ShowError(str(E))
        self.flag_valid = not flag_error
        self.__set_error_text(emsg)



# todo plot "X" instead of line
# find line in editor by clicking
# rename molecule

__all__ = ["XFileAtomsHistogram"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT # as NavigationToolbar2QT
import matplotlib.pyplot as plt
from ._shared import *
import a99


MAX_NUM_BINS = 500


class XFileAtomsHistogram(QMainWindow):
    """
    Interactive window that plots a histogram of selected field of a FileAtoms
    object.

    Args:
      file_atoms: FileAtoms object.
    """

    def __init__(self, file_atoms):
        QMainWindow.__init__(self)

        self.f = file_atoms  # FileAtoms object

        # # Central widget containing toolbar and plot area

        # ## Toolbar

        l = self.labelDataField = QLabel("&Data field")
        c = self.comboBoxDataField = QComboBox()
        c.addItems(ATOM_ATTR_NAMES)
        l.setBuddy(c)
        l2 = self.labelSpinBox = QLabel("&Bins")
        sb = self.spinBox = QSpinBox()
        sb.setMaximum(MAX_NUM_BINS)
        sb.setValue(50)
        l2.setBuddy(sb)
        b = self.puishButtonPlot = QPushButton("&Plot")
        b.clicked.connect(self.on_plot)

        s = self.spacer0 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        bb = [l, c, l2, sb, b]

        l0 = self.layoutToolbar = QHBoxLayout()
        for b in bb:
            l0.addWidget(b)
        l0.addItem(s)
        a = self.widgetPlotToolbar = QWidget()
        a.setLayout(l0)
        a.setFixedHeight(40)

        # ## Plot widget
        # http://stackoverflow.com/questions/12459811
        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.toolbar = NavigationToolbar2QT(self.canvas, self)
        l1 = self.layoutPlot = QVBoxLayout()
        l1.addWidget(self.toolbar)
        l1.addWidget(self.canvas)
        a99.set_margin(l1, 0)
        # a = self.widgetPlot = QWidget()
        # a.setLayout(l1)

        # ## Mounts central widget
        l2 = self.layoutCentral = QVBoxLayout()
        l2.addWidget(self.widgetPlotToolbar)
        l2.addLayout(l1)
        # l2.addWidget(self.widgetPlot)
        a99.set_margin(l2, 0)
        a = self.centralWidget = QWidget()
        a.setLayout(l2)
        a.setFont(a99.MONO_FONT)
        self.setCentralWidget(self.centralWidget)

        # # Final adjustments
        self.setWindowTitle("Histogram")
        a99.place_center(self)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots

    def on_plot(self, _):
        self._plot()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Gear

    def _plot(self):
        attr_name = str(self.comboBoxDataField.currentText())
        num_bins = int(self.spinBox.value())
        v = self.f.__getattribute__(attr_name)

        self.figure.clear()
        ax = self.figure.add_subplot(111)
        ax.hist(v, num_bins)
        ax.set_xlabel(attr_name)
        ax.set_ylabel("counts")
        plt.tight_layout()
        a99.format_BLB()
        self.canvas.draw()



__all__ = ["KuruczMolLine", "KuruczMolLineOld", "KuruczMolLineOld1", "FileKuruczMolecule", "FileKuruczMoleculeOld",
           "FileKuruczMoleculeBase", "load_kurucz_mol", "FileKuruczMoleculeOld1", "FileKuruczMolecule1"]

import a99
from f311 import DataFile
import io
import os
from dataclasses import dataclass
from typing import Any


# Will update progress status every time the following number of lines is read from file
_PROGRESS_INDICATOR_PERIOD = 1030 * 5


@dataclass
class KuruczMolLine:
    lambda_: Any
    loggf: Any
    J2l: Any
    E2l: Any
    Jl: Any
    El: Any
    atomn0: Any
    atomn1: Any
    state2l: Any
    v2l: Any
    lambda_doubling2l: Any
    spin2l: Any
    statel: Any
    vl: Any
    lambda_doublingl: Any
    spinl: Any
    iso: Any
    rest: Any


@dataclass
class KuruczMolLineOld:
    lambda_: Any
    J2l: Any
    Jl: Any
    atomn0: Any
    atomn1: Any
    state2l: Any
    v2l: Any
    lambda_doubling2l: Any
    spin2l: Any
    statel: Any
    vl: Any
    lambda_doublingl: Any
    spinl: Any


@dataclass
class KuruczMolLineOld1:
    lambda_: Any
    J2l: Any
    Jl: Any
    state2l: Any
    v2l: Any
    spin2l: Any
    statel: Any
    vl: Any
    spinl: Any


def load_kurucz_mol(filename):
    """
    Attempts to load Kurucz molecular line list (tries old and new format)

    Returns:
        FileKuruczMoleculeBase: either FileKuruczMolecule or FileKuruczMoleculeOld
    """
    import f311
    return f311.load_with_classes(filename, [FileKuruczMolecule, FileKuruczMolecule1,
                                           FileKuruczMoleculeOld, FileKuruczMoleculeOld1,])


class FileKuruczMoleculeBase(DataFile):
    """Base class for the two types of Kurucz molecular lines file"""

    attrs = ["num_lines"]

    @property
    def num_lines(self):
        return len(self)

    def __len__(self):
        return len(self.lines)

    def __init__(self):
        DataFile.__init__(self)

        # list of KuruczMolLine objects
        self.lines = []

    def __iter__(self):
        return iter(self.lines)

    def __getitem__(self, item):
        return self.lines.__getitem__(item)


@a99.froze_it
class FileKuruczMolecule(FileKuruczMoleculeBase):
    """
    Kurucz molecular lines file

    **Note** Kurucz file so far refers to one molecule only.

    **Note** Load only
    """

    def _do_load(self, filename):

        # **note** Kurucz puts always the "double-line" values before the "line" values
        #
        #   Wl(nm)   loggf   J"    E(cm-1)  J'   E(cm-1)   H
        #   |        |       |     |        |    |         |O
        #   |        |       |     |        |    |         || electronic state
        #   |        |       |     |        |    |         || |v"=00
        #   |        |       |     |        |    |         || || lambda-doubling component
        #   |        |       |     |        |    |         || || |spin2l
        #   |        |       |     |        |    |         || || ||
        #
        #   204.5126 -7.917  2.5    83.925  2.5  48964.990 108X00f1   A07e1   16
        #   204.7561 -7.745  3.5   202.380  3.5  49025.320 108X00f1   A07e1   16
        #   204.9400 -7.883  5.5   543.596  6.5  49322.740 108X00e1   A07e1   16
        #   205.0076 -7.931  3.5   201.931  2.5  48964.990 108X00e1   A07e1   16
        #   205.0652 -7.621  4.5   355.915  4.5  49105.280 108X00f1   A07e1   16
        #   205.0652 -7.621  4.5   355.915  4.5  49105.280 108X00f1   A07e1   16
        #
        # Another sample from file cnax12brookek.asc
        #  1500.0057 -3.734  7.5-35901.223  6.5 -42566.043 607X20f2   A22f2   12 617 K
        #  1500.0122 -1.704100.5-34572.316101.5 -41237.107 607X09 2   A10f2   12 617 K
        #  1500.0307 -2.220 84.5-33050.728 85.5 -39715.437 607X11f2   A12f2   12 617 K
        #  1500.0701 -5.367 56.5-29897.673 57.5 -36562.207 607X13f2   A14f1   12 617 K
        #  1500.0896 -3.745 95.5-22615.737 96.5  29280.184 607X03f2   A03f1   12 617 K BR
        #
        # 1         11     18   23        33   38         49          61      69 1-based
        # 0         10     17   22        32   37         48          60      68 0-based
        # 01234567890123456789012345678901234567890123456789012345678901234567890123456789
        #
        # FORMAT(F10.4.F7.3,F5.1,F10.3,F5.1,F11.3,I4,A1,I2,A1,I1,3X,A1,I2,A1,I1,3X,I2)
        # "
        # The code for the diatomic molecules is two 2-digit element numbers in
        # ascending order.  The labels consist of the electronic state, the vibrational
        # level, the lambda-doubling component, and the spin state.  Sometimes two
        # characters are required for the electronic state and the format becomes
        # ,A2,I2,A1,i1,2X,.  Negative energies are predicted or extrapolated"
        #
        # http://kurucz.harvard.edu/linelists/0linelists.readme

        filesize = os.path.getsize(filename)
        num_lines = int(filesize/70)

        with open(filename, "r") as h:
            self._do_load_h(h, filename, num_lines)

    def _do_load_h(self, h, filename, num_lines=0):
        r = 0  # counts rows of file
        ii = 0
        try:
            self.lines = []
            while True:
                s = h.readline().strip("\n")
                if len(s) == 0:
                    break

                # Kurucz: "negative energies are predicted or extrapolated" ...
                # (http: // kurucz.harvard.edu / linelists.html)
                E2l = float(s[22:32])
                # (20230720) ... but it doesn't mean that I need to invert the value: loader should be as impartial as
                #            possible
                # if E2l < 0:
                #     E2l = -E2l
                El = float(s[37:48])
                # if El < 0:
                #     El = -El

                try:
                    spin2l = int(s[56:57])
                except ValueError:
                    spin2l = 0

                try:
                    spinl = int(s[64:65])
                except ValueError:
                    spinl = 0

                line = KuruczMolLine(
                    lambda_=float(s[0:10]) * 10,
                    loggf=float(s[10:17]),
                    J2l=float(s[17:22]),
                    E2l=E2l,
                    Jl=float(s[32:37]),
                    El=El,
                    atomn0=int(s[48:50]),
                    atomn1=int(s[50:52]),
                    state2l=s[52:53],
                    v2l=int(s[53:55]),
                    lambda_doubling2l=s[55:56],
                    spin2l=spin2l,
                    statel=s[60:61],
                    vl=int(s[61:63]),
                    lambda_doublingl=s[63:64],
                    spinl=spinl,
                    iso=int(s[68:70]),
                    rest=s[70:])


                self.lines.append(line)
                r += 1
                ii += 1
                if ii == _PROGRESS_INDICATOR_PERIOD:
                    a99.get_python_logger().info(
                        "Loading '{}': {}".format(filename, a99.format_progress(r, num_lines)))
                    ii = 0


        except Exception as e:
            # f = type(e)(("Error around %d%s row of file '%s'" %
            #              (r + 1, a99.ordinal_suffix(r + 1), filename)) + ": " + str(
            #     e)).with_traceback(sys.exc_info()[2])
            raise RuntimeError("Error around %d%s row of file '%s': \"%s\"" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename, a99.str_exc(e))) from e


    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            for l in self.lines:
                #  1500.0896 -3.745 95.5-22615.737 96.5  29280.184 607X03f2   A03f1   12 617 K BR
                a99.write_lf(h, f"{l.lambda_/10:10.4f}{l.loggf:7.3f}{l.J2l:5.1f}{l.E2l:10.3f}{l.Jl:5.1f}{l.El:11.3f}"
                                f"{l.atomn0:2d}{l.atomn1:02d}{l.state2l:1s}{l.v2l:02d}{l.lambda_doubling2l:1s}"
                                f"{l.spin2l:1d}   {l.statel:1s}{l.vl:02d}{l.lambda_doublingl:1s}{l.spinl:1d}"
                                f"   {l.iso:2d}{l.rest}")


 # 15000.896 -3.745 95.5 22615.737 96.5  29280.184 607X 3f2   A 3f112 617 K BR
 # 1500.0896 -3.745 95.5-22615.737 96.5  29280.184 607X03f2   A03f1   12 617 K BR


@a99.froze_it
class FileKuruczMolecule1(FileKuruczMoleculeBase):
    """
    Kurucz molecular lines file following format of file "c2dabrookek.asc"
    """

    def _do_load(self, filename):

        # **note** Kurucz puts always the "double-line" values before the "line" values
        #
        #     Wl(nm) loggf    J"   E(cm-1)   J'    E(cm-1) C C
        #          |      |    |         |    |          | | |multiplicity
        #          |      |    |         |    |          | | ||electronic state
        #          |      |    |         |    |          | | |||v"=00
        #          |      |    |         |    |          | | |||| lambda-doubling component
        #          |      |    |         |    |          | | |||| |spin2l
        #          |      |    |         |    |          | | |||| ||
        #   287.7558-14.533 23.0  2354.082 24.0 -37095.578 6063a00e1  3d10e3  12 677  34741.495
        #   287.7564-14.955 22.0  2282.704 23.0 -37024.124 6063a00f1  3d10f3  12 677  34741.419
        #   287.7582-14.490 21.0  2214.696 22.0 -36955.900 6063a00e1  3d10e3  12 677  34741.205
        #   287.7613-15.004 24.0  2428.453 25.0 -37169.280 6063a00f1  3d10f3  12 677  34740.828
        #   287.7650-14.899 20.0  2149.765 21.0 -36890.147 6063a00f1  3d10f3  12 677  34740.382
        #   287.7671-14.573 25.0  2506.275 26.0 -37246.411 6063a00e1  3d10e3  12 677  34740.136
        #   287.7739-14.442 19.0  2088.132 20.0 -36827.442 6063a00e1  3d10e3  12 677  34739.310
        # 0         1         2         3         4         5         6         7         8
        # 012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789
        #   205.0652 -7.621  4.5   355.915  4.5  49105.280 108X00f1   A07e1   16

        filesize = os.path.getsize(filename)
        num_lines = int(filesize/84)

        with open(filename, "r") as h:
            self._do_load_h(h, filename, num_lines)

    def _do_load_h(self, h, filename, num_lines=0):
        r = 0  # counts rows of file
        ii = 0
        try:
            self.lines = []
            while True:
                s = h.readline().strip("\n")
                if len(s) == 0:
                    break

                # Kurucz: "negative energies are predicted or extrapolated"
                # (http: // kurucz.harvard.edu / linelists.html)
                E2l = float(s[22:32])
                if E2l < 0:
                    E2l = -E2l
                El = float(s[37:48])
                if El < 0:
                    El = -El

                try:
                    spin2l = int(s[57:58])
                except ValueError:
                    spin2l = 0

                try:
                    spinl = int(s[65:66])
                except ValueError:
                    spinl = 0

                line = KuruczMolLine(
                    float(s[0:10]) * 10,
                    float(s[10:17]),
                    float(s[17:22]),
                    E2l,
                    float(s[32:37]),
                    El,
                    int(s[48:50]),
                    int(s[50:52]),
                    s[53:54],
                    int(s[54:56]),
                    s[56:57],
                    spin2l,
                    s[61:62],
                    int(s[62:64]),
                    s[64:65],
                    spinl,
                    int(s[68:70]), )


                self.lines.append(line)
                r += 1
                ii += 1
                if ii == _PROGRESS_INDICATOR_PERIOD:
                    a99.get_python_logger().info(
                        "Loading '{}': {}".format(filename, a99.format_progress(r, num_lines)))
                    ii = 0


        except Exception as e:
            # f = type(e)(("Error around %d%s row of file '%s'" %
            #              (r + 1, a99.ordinal_suffix(r + 1), filename)) + ": " + str(
            #     e)).with_traceback(sys.exc_info()[2])
            raise RuntimeError("Error around %d%s row of file '%s': \"%s\"" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename, a99.str_exc(e))) from e





@a99.froze_it
class FileKuruczMoleculeOld(FileKuruczMoleculeBase):
    """
    Kurucz molecular lines file, old format #0

    This file format is Ssimilar to FileKuruczMol format, but contains less columns

    One example of file is: ATMOS/wrk4/bruno/Mole/CH/ch36.txt
    """

    def _do_load(self, filename):

        # **Note** The variable names below are the ones found in ATMOS/wrk4/bruno/Mole/CH/selech.f
        #
        #       lamb   j2   j1     v2l
        #          |    |    |      | d2l s1l
        #          |    |    |      | |   |
        #          |    |    |      | |   | v1l
        #          |    |    |      | |   | | d1l
        #          |    |    |      | |   | | |
        #   3001.028 16.5 17.5 106X02F2   C03F2
        # 0         1         2         3
        # 0123456789012345678901234567890123456


        filesize = os.path.getsize(filename)
        num_lines = int(filesize / 37)

        with open(filename, "r") as h:
            self._do_load_h(h, filename, num_lines)

    def _do_load_h(self, h, filename, num_lines=0):
        r = 0  # counts rows of file
        ii = 0
        try:
            self.lines = []
            while True:
                s = h.readline().strip("\n")
                if len(s) == 0:
                    break

                line = KuruczMolLineOld(
                    float(s[0:10]),
                    float(s[10:15]),
                    float(s[15:20]),
                    int(s[20:22]),
                    int(s[22:24]),
                    s[24:25],
                    int(s[25:27]),
                    s[27:28],
                    int(s[28:29]),
                    s[32:33],
                    int(s[33:35]),
                    s[35:36],
                    int(s[36:37]),
                )

                self.lines.append(line)
                r += 1
                ii += 1
                if ii == _PROGRESS_INDICATOR_PERIOD:
                    a99.get_python_logger().info(
                        "Loading '{}': {}".format(filename,
                                                  a99.format_progress(r, num_lines)))
                    ii = 0

        except Exception as e:
            raise RuntimeError("Error around %d%s row of file '%s': \"%s\"" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename,
                                a99.str_exc(e))) from e


@a99.froze_it
class FileKuruczMoleculeOld1(FileKuruczMoleculeBase):
    """
    Kurucz molecular lines file, old format #1

    Not sure if this is really a Kurucz format. Found one file like this in:
    ATMOS/wrk4/bruno/Mole/CN/cnbx36.txt
    """

    def _do_load(self, filename):

        # **Note** The variable names below are the ones found in ATMOS/wrk4/bruno/Mole/CN/selecn.f
        #
        #      lamb   j2l   j1l v2l d2l
        #         |     |     |   | |   v1l
        #         |     |     |   | |   | d1l
        #         |     |     |   | |   | |
        #  3601.583 149.5 150.5 X00 2 B00 2

        # 0         1         2         3
        # 0123456789012345678901234567890123456


        filesize = os.path.getsize(filename)
        num_lines = int(filesize/33)

        with open(filename, "r") as h:
            self._do_load_h(h, filename, num_lines)

    def _do_load_h(self, h, filename, num_lines=0):
        r = 0  # counts rows of file
        ii = 0
        try:
            self.lines = []
            while True:
                s = h.readline().strip("\n")
                if len(s) == 0:
                    break




                line = KuruczMolLineOld1(
                    float(s[0:9]),
                    float(s[9:15]),
                    float(s[15:21]),
                    s[22:23],
                    int(s[23:25]),
                    int(s[26:27]),
                    s[28:29],
                    int(s[29:31]),
                    int(s[32:33]),
                    )

                self.lines.append(line)
                r += 1
                ii += 1
                if ii == _PROGRESS_INDICATOR_PERIOD:
                    a99.get_python_logger().info(
                        "Loading '{}': {}".format(filename, a99.format_progress(r, num_lines)))
                    ii = 0

        except Exception as e:
            raise RuntimeError("Error around %d%s row of file '%s': \"%s\"" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename, a99.str_exc(e))) from e


def _fake_file():
    """Returns StringIO for a fragment of a Kurucz molecular lines file"""

    f = io.StringIO()
    f.write("""  204.5126 -7.917  2.5    83.925  2.5  48964.990 108X00f1   A07e1   16
  204.7561 -7.745  3.5   202.380  3.5  49025.320 108X00f1   A07e1   16
  204.9400 -7.883  5.5   543.596  6.5  49322.740 108X00e1   A07e1   16
  205.0076 -7.931  3.5   201.931  2.5  48964.990 108X00e1   A07e1   16
  205.0652 -7.621  4.5   355.915  4.5  49105.280 108X00f1   A07e1   16
""")

    f.seek(0)
    return f


def _test():
    """Test code in docstring

    Example:

    >>> h = _fake_file()
    >>> f = FileKuruczMolecule()
    >>> f._do_load_h(h, "_fake_file")
    >>> f.lines[0]
    KuruczMolLine(2045.126, -7.917, 2.5, 83.925, 2.5, 48964.99, 1, 8, 'X', 0, 'f', 1, 'A', 7, 'e', 1, 6)
    """
    return




"""Widget to edit a FileAbonds object."""

__all__ = ["WFileAbonds"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
import pyfant


ABONDS_HEADERS = ["Element", "Abundance", "Notes"]
NOTES_COLUMN_WIDTH = 200


def _format_error(row_index, message):
    return "<b>Row %d</b>: %s" % (row_index+1, message)


class WFileAbonds(QWidget):
    """
    FileAbonds editor widget.

    Args:
      parent=None
    """

    # Emitted whenever any value changes
    changed = pyqtSignal()
    # Emitted whenever a new file is loaded
    loaded = pyqtSignal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        # Whether all the values in the fields are valid or not
        self.flag_valid = False
        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False
        self.f = None # FileAbonds object

        # Internals
        d = self.__default_dissoc = pyfant.FileDissoc()
        d.init_default()

        la = self.formLayout = QVBoxLayout()
        a99.set_margin(la, 0)
        la.setSpacing(4)

        self.setLayout(la)


        # # Sort/Insert/append/delete toolbar

        l = self.c29378 = QHBoxLayout()
        la.addLayout(l)
        a99.set_margin(l, 0)
        l.setSpacing(4)
        b = self.button_sort_a = QPushButton("Sort &alphabetically")
        l.addWidget(b)
        b.clicked.connect(self.on_sort_a)
        b = self.button_sort_z = QPushButton("Sort by atomic &number")
        l.addWidget(b)
        b.clicked.connect(self.on_sort_z)
        l.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        l = self.c34985 = QHBoxLayout()
        la.addLayout(l)
        a99.set_margin(l, 0)
        l.setSpacing(4)
        b = self.button_insert = QPushButton("&Insert")
        l.addWidget(b)
        b.clicked.connect(self.on_insert)
        b = self.button_append = QPushButton("A&ppend")
        l.addWidget(b)
        b.clicked.connect(self.on_append)
        b = self.button_delete = QPushButton("&Delete")
        l.addWidget(b)
        b.clicked.connect(self.on_delete)
        l.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        # # Splitter notes box; the table; and the errors area

        sp = self.splitter = QSplitter(Qt.Vertical)
        la.addWidget(sp)

        # # Edit box for the notes

        w = self.c3f1x2 = QWidget()
        sp.addWidget(w)
        l = self.c114_2 = QHBoxLayout(w)
        a99.set_margin(l, 0)
        l.setSpacing(4)
        label = self.c33_1 = QLabel("&Notes")
        l.addWidget(label)
        edit = self.textEditNotes = QPlainTextEdit()
        edit.textChanged.connect(self.on_textEditNotes_textChanged)
        l.addWidget(edit)
        label.setBuddy(edit)
        T = "Use this space to write annotations such as references"
        label.setToolTip(T)
        edit.setToolTip(T)


        # ## The table widget

        a = self.tableWidget = QTableWidget()
        sp.addWidget(a)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        # a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)
        a.cellChanged.connect(self.on_tableWidget_cellChanged)
        a.setEditTriggers(QAbstractItemView.DoubleClicked | QAbstractItemView.EditKeyPressed)
        a.setFont(a99.MONO_FONT)
        a.installEventFilter(self)

        # ## The errors area

        w = self.csslff = QWidget()
        sp.addWidget(w)
        l = self.c49378 = QVBoxLayout(w)
        a99.set_margin(l, 0)
        l.setSpacing(1)

        x = self.c88888 = QLabel("<b>Errors</b>")
        l.addWidget(x)

        x = self.textEditError = QTextEdit(self)
        l.addWidget(x)
        x.setReadOnly(True)
        x.setStyleSheet("QTextEdit {color: %s}" % a99.COLOR_ERROR)
        x.setFont(a99.MONO_FONT)

        # ## Splitter stretch factors
        # These need to be set a posteriori otherwise they do
        # not work properly.
        sp.setStretchFactor(0, 0)
        sp.setStretchFactor(1, 1)
        sp.setStretchFactor(2, 0)

        # finally...
        self.setEnabled(False)  # Disabled until load() is called
        self.flag_process_changes = True


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def load(self, x):
        assert isinstance(x, pyfant.FileAbonds)
        self.f = x
        self._update_from_file_abonds()
        # this is called to perform file validation upon loading
        self._update_file_abonds()
        self.setEnabled(True)
        self.loaded.emit()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def showEvent(self, event):
        self.setFocus()

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""
        self.tableWidget.setFocus()

    def eventFilter(self, source, event):
        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Return:
                if source == self.tableWidget:
                    self.tableWidget.editItem(self.tableWidget.currentItem())
                    return True
        return False


    # def Validate(self):
    #     # Currently calling Parameters.GetKwargs() to do the <quote>validation<quote>
    #     self._parameters.UpdateFromWidgets()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_tableWidget_cellChanged(self, row, column):
        if self.flag_process_changes:
            self.flag_process_changes = False
            try:
                t = self.tableWidget
                item = t.item(row, column)
                flag_done = False
                if column == 0:  # Element
                    try:
                        item.setText(self._validate_element(row, item.text()))
                        flag_done = True
                    except Exception as E:
                        a99.show_error(str(E))

                if flag_done:
                    self._update_file_abonds()


            finally:
                self.flag_process_changes = True

            self._update_file_abonds()
            self.changed.emit()

    def on_sort_a(self):
        self.f.sort_a()
        self._update_from_file_abonds()
        self._update_file_abonds()
        self.changed.emit()

    def on_sort_z(self):
        not_found = self.f.sort_z()
        self._update_from_file_abonds()
        self._update_file_abonds()
        self.changed.emit()
        if len(not_found) > 0:
            a99.show_message("Symbols not found in the periodic table:\n\n"+
                            str([x.strip() for x in not_found])+"\n\n"+
                            "These symbols will appear first and will be ordered alphabetically.")

    def on_insert(self):
        self._insertRow(self.tableWidget.currentRow())

    def on_append(self):
        self._insertRow(self.tableWidget.rowCount())

    def on_delete(self):
        r = QMessageBox.question(self, "Delete row", "Are you sure?",
         QMessageBox.Yes|QMessageBox.No, QMessageBox.Yes)
        if r == QMessageBox.Yes:
            self.flag_process_changes = False
            try:
                t = self.tableWidget
                t.removeRow(t.currentRow())
                self._update_file_abonds()
            finally:
                self.flag_process_changes = True
            self.changed.emit()

    def on_textEditNotes_textChanged(self):
        if self.flag_process_changes:
            self._update_file_abonds()
            self.changed.emit()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear

    def _insertRow(self, pos):
        self.flag_process_changes = False
        try:
            t = self.tableWidget
            t.insertRow(pos)
            item = QTableWidgetItem("??")
            t.setItem(pos, 0, item)
            item = QTableWidgetItem("0.0")
            t.setItem(pos, 1, item)
            item = QTableWidgetItem("")
            t.setItem(pos, 2, item)
            t.setCurrentCell(pos, 0)
            self._update_file_abonds()
        finally:
            self.flag_process_changes = True
        self.changed.emit()

    def _set_error_text(self, x):
        """Sets text of textEditError."""
        self.textEditError.setText(x)

    def _update_from_file_abonds(self):
        self.flag_process_changes = False
        try:
            o, t = self.f, self.tableWidget
            n = len(o)
            a99.reset_table_widget(t, n, len(ABONDS_HEADERS))
            t.setHorizontalHeaderLabels(ABONDS_HEADERS)

            # list with the vectors themselves

            attrs = [o.__getattribute__(x) for x in ["ele", "abol", "notes_per_ele"]]

            for i in range(len(o)):
                for j, attr in enumerate(attrs):
                    item = QTableWidgetItem(str(attr[i]).strip())
                    t.setItem(i, j, item)
            t.resizeColumnsToContents()
            t.setColumnWidth(2, NOTES_COLUMN_WIDTH)  # Make room for notes for element

            self.textEditNotes.setPlainText(o.notes)
        finally:
            self.flag_process_changes = True

    def _validate_element(self, row_index, x):
        """Validates element symbol and returns it stripped and uppercase.
        Raises in cases of validation error."""
        x = str(x).upper().strip()
        if len(x) > 2:
            raise RuntimeError(
                "Element \"%s\" too big (maximum 2 characters" % x)

        # hydrogen not allowed
        if x == "H":
            raise RuntimeError("Hydrogen not allowed (abundance of H is internally fixed as =12)")

        # makes sure elements symbols are unique in the table
        t = self.tableWidget
        n = t.rowCount()
        for ii in range(n):
            if ii != row_index:
                if x == str(t.item(ii, 0).text()):
                    raise RuntimeError("Element \"%s\" already exists" % x)
        return x


    def _update_file_abonds(self):
        errors, warnings = [], []
        o, t = self.f, self.tableWidget
        assert isinstance(t, QTableWidget)
        n = t.rowCount()
        ele, abol, notes_per_ele = [], [], []

        for i in range(n):
            # # Element
            item = t.item(i, 0)
            x = str(item.text()).upper().strip()
            try:
                x = self._validate_element(i, x)
            except Exception as E:
                errors.append(_format_error(i, str(E)))
            ele.append(pyfant.adjust_atomic_symbol(x))
            item.setText(x)

            # # Abundance
            item = t.item(i, 1)
            x = 0
            try:
                x = float((item.text()))
            except Exception as E:
                errors.append(_format_error(i, str(E)))
            abol.append(x)
            item.setText(str(x))

            # # Notes per element (no validation required)
            item = t.item(i, 2)
            x = str((item.text())).strip()
            notes_per_ele.append(x)

        d = self.__default_dissoc
        for elem, cclog in zip(d.elems, d.cclog):
            if elem not in ele and elem != " H":
                warnings.append("<b>Warning</b>: element \"%s\", required for "
                 "dissociative equilibrium calculation, is missing. "
                 "Abundance adopted for \"%s\" will be %g" %
                 (elem.strip(), elem.strip(), cclog+12))

        o.ele = ele
        o.abol = abol
        o.notes_per_ele = notes_per_ele
        o.notes = self.textEditNotes.toPlainText()
        self.flag_valid = len(errors) == 0
        emsg = ""
        if len(errors) > 0 or len(warnings) > 0:
            emsg = '<ul><li>'+("<li>".join(warnings+errors))

        self._set_error_text(emsg)



"""
VALD3-to-PFANT conversions
"""

__all__ = ["vald3_to_atoms"]

import csv
import a99
import sys
import pyfant


# Here is a sample of a VALD3 file:
#  1|                                                                   Lande factors      Damping parameters
#  2|Elm Ion      WL_air(A)   log gf* E_low(eV) J lo  E_up(eV) J up  lower  upper   mean   Rad.  Stark  Waals
#  3|'OH 1',       10000.092,  -7.089,  0.0750,  4.5,  1.3150,  3.5,99.000,99.000,99.000, 0.000, 0.000, 0.000,
#  4|'  Hb                                                      2s2.3s2.1p3          X,2,1,e,5,0'
#  5|'  Hb                                                  2s2.3s2.1p3              X,2,1,e,4,3'
#  6|'GSGCD             OH data from Kur   1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD   (16)OH        '
#  7|'OH 1',       10000.117, -10.110,  2.5930, 26.5,  3.8330, 26.5,99.000,99.000,99.000, 0.000, 0.000, 0.000,
#  8|'  Hb                                                     2s2.3s2.1p3          X,2,1,f,26,3'
#  9|'  Hb                                                 2s2.3s2.1p3              X,2,1,f,27,7'
# 10|'GSGCD             OH data from Kur   1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD     1 GSGCD   (16)OH        '
# 11|'Ar 2',       10000.184,  -3.150, 25.4140,  2.5, 26.6540,  3.5,99.000,99.000,99.000, 0.000, 0.000, 0.000,
# 12|'  JK                                                           3s2.3p4.(3P<2>).5f     2[3]'
# 13|'  JK                                                           3s2.3p4.(3P<1>).7g     2[3]'
# 14|'KP                Li 1 - K 5 Bell    2 KP        2 KP        2 KP        2 KP        2 KP        2 KP        2 KP        2 KP        2 KP      Ar+           '


# Symbols of elements with first letter uppercase and second letter (if present) lowercase
_Symbols = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S',
            'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
            'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd',
            'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
            'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os',
            'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
            'Pa', 'U']


def vald3_to_atoms(file_obj):
    """
    Converts data from a VALD3 file into a FileAtoms object.

    Args:
      file_obj: file-like object, e.g., returned by open()

    Returns: a FileAtoms object

    VALD3 website: http://vald.astro.uu.se/
    """

    _logger = a99.get_python_logger()

    def log_skipping(r, reason, row):
        _logger.info("Skipping row #%d (%s)" % (r, reason))
        _logger.info(str(row))

    reader = csv.reader(file_obj)
    ret = pyfant.FileAtoms()
    edict = {}  # links atomic symbols with Atom objects created (key is atomic symbol)
    r = 0
    num_skip_ioni, num_skip_mol = 0, 0
    try:
        for row in reader:
            r += 1
            if len(row) <=  12:  # Condition to detect row of interest
                continue

            elem = row[0][1:row[0].index(" ")]
            if not elem in _Symbols:
                # skips molecule
                num_skip_mol += 1
                continue

            # # Collects information and creates atomic line object.
            # Note: Variable names follow Fortran source variable names.

            s_ioni = row[0][-2]
            ioni = int(s_ioni)  # not used, stays as file validation

            if ioni > 2:
                # log_skipping(r, "ionization > 2", row)
                num_skip_ioni += 1
                continue

            _waals = float(row[12])
            if _waals > 0:
                # If the van der Waals damping parameter is > 0, it means
                # something else, see the documentation at
                # http://www.astro.uu.se/valdwiki/Vald3Format:
                #  "if >0 : two parameters as defined in Barklem et al.
                #   2000A&AS..142..467B : integer part is cross-section sigma,
                #   fractional part is velocity parameter alpha"
                # Therefore, we fall back.
                _waals = 0

            line = pyfant.AtomicLine()
            line.lambda_ = float(row[1])
            line.algf = float(row[2])
            line.kiex = float(row[3])
            if _waals == 0:
                line.ch = 0.3e-31
            else:
                try:
                    # Formula supplied by Elvis Cantelli:
                    # extracted from cross-entropy code by P. Barklem
                    line.ch = 10**(2.5*_waals-12.32)
                except:
                    # Catches error to log _waals value that gave rise to error
                    _logger.critical("Error calculating ch: waals=%g" % _waals)
                    raise

            # Setting gr to zero will cause PFANT to calculate it using a formula.
            # See pfantlib.f90::read_atoms() for the formula.
            line.gr = 0.0
            # ge is not present in VALD3 file.
            # it enters as a multiplicative term in popadelh(). The original
            # atom4070g.dat and atoms.dat all had ge=0 for all lines.
            line.ge = 0.0
            # Attention: zinf must be tuned later using tune-zinf.py
            line.zinf = 0.5
            # Never used in PFANT
            line.abondr = 1

            # # Stores line in object
            elem = pyfant.adjust_atomic_symbol(elem)
            key = elem+s_ioni  # will group elements by this key
            if key in edict:
                a = edict[key]
            else:
                a = edict[key] = pyfant.Atom()
                a.elem = elem
                a.ioni = int(s_ioni)
                ret.atoms.append(a)
            a.lines.append(line)
    except Exception as e:
        raise type(e)(("Error around %d%s row of VALD3 file" %
            (r, a99.ordinal_suffix(r)))+": "+str(e)).with_traceback(sys.exc_info()[2])
    _logger.debug("VALD3-to-atoms conversion successful!")
    _logger.info("Number of lines skipped (molecules): %d" % num_skip_mol)
    _logger.info("Number of lines skipped (ioni > 2): %d" % num_skip_ioni)
    _logger.debug("Number of (element+ioni): %s" % len(ret))
    _logger.debug("Total number of atomic lines: %d" % (sum(len(a) for a in ret.atoms),))
    return ret





# ABANDONED, CRASHING FORM ON PLOT, decided to obtain histogram programatically only

# __all__ = ["XFileMoleculeHistogram"]
#
# from PyQt5.QtCore import *
# from PyQt5.QtGui import *
# from PyQt5.QtWidgets import *
# from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
# from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT # as NavigationToolbar2QT
# import matplotlib.pyplot as plt
# from ._shared import *
# import a99
#
#
# MAX_NUM_BINS = 500
#
# class XFileMoleculeHistogram(QMainWindow):
#     """
#     Interactive window that plots a histogram of selected field of a FileAtoms
#     object.
#
#     Args:
#       mol: pyfant.Molecule
#     """
#
#     def __init__(self, mol):
#         import pyfant
#
#         QMainWindow.__init__(self)
#
#         # self.setWindowModality(Qt.ApplicationModal)
#
#         assert isinstance(mol, pyfant.Molecule)
#         self.mol = mol
#
#         # # Central widget containing toolbar and plot area
#
#         # ## Toolbar
#
#         l = self.labelDataField = QLabel("&Data field")
#         c = self.comboBoxDataField = QComboBox()
#         c.addItems(["lmbdam", "sj", "jj"])
#         l.setBuddy(c)
#         l2 = self.labelSpinBox = QLabel("&Bins")
#         sb = self.spinBox = QSpinBox()
#         sb.setMaximum(MAX_NUM_BINS)
#         sb.setValue(50)
#         l2.setBuddy(sb)
#         b = self.puishButtonPlot = QPushButton("&Plot")
#         b.clicked.connect(self.on_plot)
#
#         s = self.spacer0 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)
#
#         bb = [l, c, l2, sb, b]
#
#         l0 = self.layoutToolbar = QHBoxLayout()
#         for b in bb:
#             l0.addWidget(b)
#         l0.addItem(s)
#         a = self.widgetPlotToolbar = QWidget()
#         a.setLayout(l0)
#         a.setFixedHeight(40)
#
#         # ## Plot widget
#         # http://stackoverflow.com/questions/12459811
#         self.figure = plt.figure()
#         self.canvas = FigureCanvas(self.figure)
#         self.toolbar = NavigationToolbar2QT(self.canvas, self)
#         l1 = self.layoutPlot = QVBoxLayout()
#         l1.addWidget(self.toolbar)
#         l1.addWidget(self.canvas)
#         a99.set_margin(l1, 0)
#         # a = self.widgetPlot = QWidget()
#         # a.setLayout(l1)
#
#         # ## Mounts central widget
#         l2 = self.layoutCentral = QVBoxLayout()
#         l2.addWidget(self.widgetPlotToolbar)
#         l2.addLayout(l1)
#         # l2.addWidget(self.widgetPlot)
#         a99.set_margin(l2, 0)
#         a = self.centralWidget = QWidget()
#         a.setLayout(l2)
#         a.setFont(a99.MONO_FONT)
#         self.setCentralWidget(self.centralWidget)
#
#         # # Final adjustments
#         self.setWindowTitle("Histogram")
#         a99.place_center(self)
#
#
#     # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
#     # Slots
#
#     def on_plot(self, _):
#         self._plot()
#
#     # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
#     # Gear
#
#     def _plot(self):
#         attr_name = str(self.comboBoxDataField.currentText())
#         num_bins = int(self.spinBox.value())
#         v = self.mol.__getattribute__(attr_name)
#
#         self.figure.clear()
#         plt.figure(self.figure)
#         ax = self.figure.add_subplot(111)
#         ax.hist(v, num_bins)
#         ax.set_xlabel(attr_name)
#         ax.set_ylabel("counts")
#         plt.tight_layout()
#         a99.format_BLB()
#         self.canvas.draw()



"""
TurboSpectrum-to-PFANT conversions
"""

__all__ = ["turbospectrum_to_atoms"]

import csv
import a99
import sys
import pyfant

# Comparison of same atomic line
#
#--|PFANT begin|--
#LI1   4601.843
#1.8479 -5.804 3e-32 0 0 0.1 1 0
#(kiex algf ch gr ge zinf abondr finrai)
#--|PFANT end|--
#
#--|TurboSpectrum begin|--
#' 3.000              '    1       110
#'Li I '
#  4601.843  1.848  -5.804  2.50    6.0  5.01E+07 'p' 'f'   0.0    1.0 'Li I    2p  2P    4f  2F     SEN    '
# 16814.155  4.541  -2.790 -5.54    2.0  3.02e+07 'x' 'x' 0.0 1.0 'LI  I   4d  2D      11p  2P   '

#--|TurboSpectrum end|--

_Symbols = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S',
            'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
            'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd',
            'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
            'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os',
            'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
            'Pa', 'U']

# File reader states
_X_EL0 = 0  # expecting first line of element specification
_X_EL1 = 1  # expecting second line of element specification
_X_LINE_OR_EL0 = 2  # expecting line data


def turbospectrum_to_atoms(file_obj):
    """
    Converts data from a TurboSpectrum file into a FileAtoms object.

    Args:
      file_obj: file-like object, e.g., returned by open()

    Returns: a FileAtoms object
    """

    _logger = a99.get_python_logger()

    def log_skipping(r, reason, row):
        _logger.info("Skipping row #%d (%s)" % (r, reason))
        _logger.info(str(row))

    ret = pyfant.FileAtoms()
    edict = {}  # links atomic symbols with Atom objects created (key is atomic symbol + ionization)
    r = 0  # number of current row
    num_skip_ioni, num_skip_mol = 0, 0
    state = _X_EL0
    try:
        for row in file_obj:
            r += 1

            if state < _X_LINE_OR_EL0:
                if row[0] != "'":
                    raise ParseError("Expecting element specification, but found something else")

            if state == _X_EL0:
                #' 3.000              '    1       110
                #0123456
                elem = _Symbols[int(row[1:3])]
                ioni = int(row[4:7])+1
                state = _X_EL1
            elif state == _X_EL1:
                state = _X_LINE_OR_EL0
            elif state == _X_LINE_OR_EL0:
                if row[0] == "'":
                    elem = _Symbols[int(row[1:3])]
                    ioni = int(row[4:7]) + 1
                    state = _X_EL1
                else:
                    if ioni > 2:
                        # log_skipping(r, "ionization > 2", row)
                        num_skip_ioni += 1
                        continue

                    #'Li I '
                    #  4601.843  1.848  -5.804  2.50    6.0  5.01E+07 'p' 'f'   0.0    1.0 'Li I    2p  2P    4f  2F     SEN    '
                    #0         1         2         3
                    #01234567890123456789012345678901234567890

                    rr = row.split()

                    _waals = float(rr[3])
                    if _waals > 0:
                        _waals = 0

                    line = pyfant.AtomicLine()
                    line.lambda_ = float(rr[0])
                    line.kiex = float(rr[1])
                    line.algf = float(rr[2])
                    if _waals == 0:
                        line.ch = 0.3e-31
                    else:
                        try:
                            # Formula supplied by Elvis Cantelli:
                            # extracted from cross-entropy code by P. Barklem
                            line.ch = 10**(2.5*_waals-12.32)
                        except:
                            # Catches error to log _waals value that gave rise to error
                            _logger.critical("Error calculating ch: waals=%g" % _waals)
                            raise

                    # Setting gr to zero will cause PFANT to calculate it using a formula.
                    # See pfantlib.f90::read_atoms() for the formula.
                    line.gr = 0.0
                    # ge is not present in VALD3 file.
                    # it enters as a multiplicative term in popadelh(). The original
                    # atom4070g.dat and atoms.dat all had ge=0 for all lines.
                    line.ge = 0.0
                    # Attention: zinf must be tuned later using tune-zinf.py
                    line.zinf = 0.5
                    # Never used in PFANT
                    line.abondr = 1

                    key = f"{elem:2}{ioni:1}"  # will group elements by this key
                    if key in edict:
                        a = edict[key]
                    else:
                        a = edict[key] = pyfant.Atom()
                        a.elem = pyfant.adjust_atomic_symbol(elem)
                        a.ioni = ioni
                        ret.atoms.append(a)
                    a.lines.append(line)

    except Exception as e:
        raise type(e)(("Error around %d%s row of TurboSpectrum file" %
            (r, a99.ordinal_suffix(r)))+": "+str(e)).with_traceback(sys.exc_info()[2])
    _logger.debug("VALD3-to-atoms conversion successful!")
    _logger.info("Number of lines skipped (molecules): %d" % num_skip_mol)
    _logger.info("Number of lines skipped (ioni > 2): %d" % num_skip_ioni)
    _logger.debug("Number of (element+ioni): %s" % len(ret))
    _logger.debug("Total number of atomic lines: %d" % (sum(len(a) for a in ret.atoms),))
    return ret







__all__ = ["FileAtoms", "Atom", "AtomicLine"]

import sys
import numpy as np
import tabulate
import a99
from f311 import DataFile
import pyfant


@a99.froze_it
class Atom(a99.AttrsPart):
    """
    Element with its atomic lines

    Atom is identified by key symbol+ionization
    """
    attrs = ["elem", "ioni"]

    @property
    def llzero(self):
        """Minimum wavelength"""
        return min([x.lambda_ for x in self.lines])

    @property
    def llfin(self):
        """Maximum wavelength"""
        return max([x.lambda_ for x in self.lines])

    # Properties that iterate through the AtomicLine objects to mount vectors
    @property
    def lambda_(self):
        return np.array([x.lambda_ for x in self.lines])
    @property
    def kiex(self):
        return np.array([x.kiex for x in self.lines])
    @property
    def algf(self):
        return np.array([x.algf for x in self.lines])
    @property
    def ch(self):
        return np.array([x.ch for x in self.lines])
    @property
    def gr(self):
        return np.array([x.gr for x in self.lines])
    @property
    def ge(self):
        return np.array([x.ge for x in self.lines])
    @property
    def zinf(self):
        return np.array([x.zinf for x in self.lines])
    @property
    def abondr(self):
        return np.array([x.abondr for x in self.lines])

    def __init__(self):
        a99.AttrsPart.__init__(self)
        self.lines = []  # list of AtomicLine
        self.elem = None
        self.ioni = None

    def __len__(self):
        return len(self.lines)

    def __str__(self):
        return "'{} {}' ({} lines)".format(self.elem, self.ioni, len(self))

    def one_liner_str(self):
        return self.__str__()

    def __repr__(self):
        return "'%s%s'" % (self.elem, self.ioni)

    def _cut(self, llzero, llfin):
        """Keeps only the lines with their llzero <= lambda_ <= llfin."""
        for i in reversed(list(range(len(self)))):
            if not (llzero <= self.lines[i].lambda_ <= llfin):
                del self.lines[i]


@a99.froze_it
class AtomicLine(a99.AttrsPart):
    attrs = ["lambda_", "kiex", "algf", "ch", "gr", "ge", "zinf", "abondr"]

    def __init__(self):
        a99.AttrsPart.__init__(self)

        # all scalars
        self.lambda_ = None
        self.kiex = None
        self.algf = None
        self.ch = None
        self.gr = None
        self.ge = None
        self.zinf = None
        self.abondr = None

    def filter(self, lzero, lfin):
        """Reduces the number of lines to only the ones whose lambda is inside [lzero, lfin]"""
        raise NotImplementedError()


class FileAtoms(DataFile):
    """
    PFANT Atomic Lines
    """

    default_filename = "atoms.dat"
    attrs = ["atoms", "num_lines"]
    editors = ["ated.py"]

    @property
    def llzero(self):
        """Minimum wavelength"""
        return min([a.llzero for a in self.atoms])

    @property
    def llfin(self):
        """Maximum wavelength"""
        return max([a.llfin for a in self.atoms])

    @property
    def num_lines(self):
        ret = sum(map(len, self.atoms))
        return ret

    # Properties that iterate through the AtomicLine objects to mount vectors
    @property
    def lines(self):
        """Merges all AtomicLine objects. Returns a list."""
        ret = []
        for a in self.atoms:
            ret.extend(a.lines)
        return ret
    @property
    def lambda_(self):
        return np.hstack([np.array([x.lambda_ for x in a.lines]) for a in self.atoms])
    @property
    def kiex(self):
        return np.hstack([np.array([x.kiex for x in a.lines]) for a in self.atoms])
    @property
    def algf(self):
        return np.hstack([np.array([x.algf for x in a.lines]) for a in self.atoms])
    @property
    def ch(self):
        return np.hstack([np.array([x.ch for x in a.lines]) for a in self.atoms])
    @property
    def gr(self):
        return np.hstack([np.array([x.gr for x in a.lines]) for a in self.atoms])
    @property
    def ge(self):
        return np.hstack([np.array([x.ge for x in a.lines]) for a in self.atoms])
    @property
    def zinf(self):
        return np.hstack([np.array([x.zinf for x in a.lines]) for a in self.atoms])
    @property
    def abondr(self):
        return np.hstack([np.array([x.abondr for x in a.lines]) for a in self.atoms])

    def __len__(self):
        """Length of FileAtoms object is defined as number of elements."""
        return len(self.atoms)

    def __init__(self):
        DataFile.__init__(self)

        # list of Atom objects
        self.atoms = []

    def __str__(self):
        data = [["{} {}".format(atom.elem, atom.ioni), len(atom)] for atom in self]
        headers = ["Species", "Number of lines"]
        return tabulate.tabulate(data, headers)

    def find_atom(self, search):
        """Returns Atom object given search string such as "Fe1"; may raise ValueError or IndexError"""
        elem, ioni = pyfant.str_to_elem_ioni(search)

        for atom in self.atoms:
            if atom.elem == elem and atom.ioni == ioni:
                return atom

        raise IndexError("Cannot find '{}'".format(search))


    def cut(self, llzero, llfin):
        """Keeps only the lines with their llzero <= lambda_ <= llfin."""
        for i in reversed(list(range(len(self)))):
            atom = self.atoms[i]
            atom._cut(llzero, llfin)
            if len(atom) == 0:
                del self.atoms[i]

    def filter(self, function):
        """
        Filters atomic lines for which function(line) is true.

        Args:
            function: callable:

                - receives an AtomicLine object as argument,
                - must return True or False (meaning whether or not you want that atomic line)

        "function" example: ``lambda line: line.algf >= -7``
        """
        for i in reversed(list(range(len(self)))):
            atom = self.atoms[i]
            atom.lines = list(filter(function, atom.lines))
            if len(atom) == 0:
                del self.atoms[i]

    filter_lines = filter

    def filter_atoms(self, function):
        """
        Filters Atom objects for which function(line) is true.

        Args:
          function: receives an Atomic object as argument.
           Example: lambda atom: atom.ioni <= 2
        """
        self.atoms = list(filter(function, self.atoms))

    def remove_element(self, elem):
        """Removes given element (any ionization level)."""
        elem = pyfant.adjust_atomic_symbol(elem)
        for i in reversed(list(range(len(self)))):
            atom = self.atoms[i]
            if atom.elem == elem:
                del self.atoms[i]


    def __iter__(self):
        return iter(self.atoms)

    def _do_load(self, filename):
        """Clears internal lists and loads from file."""

        with open(filename, "r") as h:

            r = 0 # counts rows of file
            edict = {}  # links atomic symbols with Atom objects created (key is atomic symbol)
            try:
                while True:
                    line = AtomicLine()

                    # (EE)(I) --whitespace-- (float) --ignored...--
                    temp = a99.str_vector(h)
                    elem, s_ioni = temp[0][:-1], temp[0][-1]
                    line.lambda_ = float(temp[1])
                    elem = pyfant.adjust_atomic_symbol(elem)
                    key = elem+s_ioni  # will gb.py elements by this key
                    if key in edict:
                        a = edict[key]
                    else:
                        a = edict[key] = Atom()
                        a.elem = elem
                        a.ioni = int(s_ioni)
                        self.atoms.append(a)
                    a.lines.append(line)
                    r += 1

                    [line.kiex, line.algf, line.ch, line.gr, line.ge, line.zinf, line.abondr, finrai] = \
                        a99.float_vector(h)
                    r += 1
                    # last element is end-of-file flag "finrai"
                    if finrai == 1:
                        break
            except Exception as e:
                raise type(e)(("Error around %d%s row of file '%s'" %
                               (r+1, a99.ordinal_suffix(r+1), filename))+": "+str(e)).with_traceback(sys.exc_info()[2])

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            n = len(self.atoms)
            for i, e in enumerate(self.atoms):
                p = len(e)
                for j, a in enumerate(e.lines):
                    finrai = 1 if i == n-1 and j == p-1 else 0
                    a99.write_lf(h, "%2s%1d %10.3f" % (e.elem, e.ioni, a.lambda_))
                    # Writing floating-point numbers with %.7g format creates a more compact
                    # file. In Fortran it is read with '*' format, so it understands as
                    # long as there is a space between numbers.
                    a99.write_lf(h, "%.7g %.7g %.7g %.7g %.7g %.7g %.7g %1d" % \
                             (a.kiex, a.algf, a.ch, a.gr, a.ge, a.zinf, a.abondr, finrai))
                    # old way write_lf(h, "%8.3f %.7g %8.3f %8.3f %8.3f %6.1f %3.1f %1d" % \
                    #     (a.kiex, a.algf, a.ch, a.gr, a.ge, a.zinf, a.abondr, finrai))



#!/usr/bin/env python

import a99
import os
import sys

# TODO not working

if __name__ == "__main__":
    path_ = a99.get_path_to_db("hitrandb")
    if os.path.isfile(path_):
        while True:
            ans = input("File '{}' already exists, overwrite it (Y/n)? ".format(path_)).upper()
            if ans in ("N", "NO"):
                sys.exit()
            if ans in ("Y", "YES", ""):
                break
        os.unlink(path_)

    db.create_db()
    db.populate_db()
    a99.get_python_logger().info("Successfully created file '{}'".format(path_))



import a99
import logging

# Tests importing packages robobrowser and bs4, requirements for downloaders
flag_ok = True
try:
    try:
        import werkzeug
        werkzeug.cached_property = werkzeug.utils.cached_property
    except AttributeError:
        pass
    from robobrowser import RoboBrowser
    del RoboBrowser
except ImportError as e:
    flag_ok = False
    logging.warning("failed to import package robobrowser.Robobrowser: '{}', "
                    "f311.filetypes.downloaders will not be available".format(a99.str_exc(e)))

if flag_ok:
    try:
        import bs4
        del bs4
    except ImportError as e:
        flag_ok = False
        logging.warning("failed to import package bs4 (Beautiful Soup 4): '{}', "
                        "f311.filetypes.downloaders will not be available".format(a99.str_exc(e)))

del a99, logging

if flag_ok:
    from .downhitran import *
    from .downnist import *
